import os
import re
import glob
import argparse
import json
import enum
import collections
from itertools import product, chain, repeat
from multiprocessing import Pool

import numpy as np
from scipy import stats
import tqdm

from network_configs import StaticModel, TemporalModel, ProbDist

class welfords:
  def __init__(self):
    self.__count = 0
    self.__m = None
    self.__s = None

  def update(self, new_val):
    if self.__count == 0:
      self.__m = np.zeros_like(new_val, dtype=np.float)
      self.__s = np.zeros_like(new_val, dtype=np.float)

    self.__count += 1
    new_m = self.__m + (new_val - self.__m)*1.0/self.__count
    new_s = self.__s + (new_val - self.__m)*(new_val - new_m)

    self.__m = new_m
    self.__s = new_s

  def count(self):
    return self.__count

  def sum(self):
    return self.__count * self.__m

  def mean(self):
    return self.__m

  def std(self):
    if self.__count < 2:
      return np.full_like(self.__m, np.nan, dtype=np.float)
    return np.sqrt(self.__s/(self.__count - 1))

  def sem(self):
    return self.std()/np.sqrt(self.__count)

def to_array(num_array):
  return np.array(num_array, dtype=np.float)

class NumpyEncoder(json.JSONEncoder):
  def default(self, obj):
   if type(obj) == float and np.isnan(val):
     return None
   if isinstance(obj, np.ndarray):
     return obj.tolist()
   return json.JSONEncoder.default(self, obj)

def load_data(input_prefix, output_prefix, static_model, temporal_model, deg_id):
  print("now aggregating:", static_model, temporal_model, deg_id)

  single_required = [
      "dt", "seed", "window",
      "network", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "lifetime-dist",
      "mean-mass", "mean-volume", "mean-lifetime", "mass-times-dist", "volume-times-dist",
      "mass-times-to-end-dist", "volume-times-to-end-dist",
      "censored-event-times-dist", "censored-lifetime-dist"]

  homogeneous_required = [
      "dt", "seed", "window",
      "network", "field", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "occupied-event-times-dist",
      "occupied-event-times-to-end-dist"]

  homogeneous_alt_required = [
      "dt", "seed", "window",
      "network", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "dt-band-homogeneous-occupied", "dt-band-external-field-occupied",
      "homogeneous-occupied-count-dist", "homogeneous-occupied-count-to-end-dist",
      "external-field-occupied-count-dist", "external-field-occupied-count-to-end-dist"]

  dist_required = [
      "dt", "seed", "counts", "network"]

  network_required = ["temporal-edges", "temporal-nodes"]

  single_summaries = {}
  single_files = glob.glob(os.path.join(
                        input_prefix, static_model.value,
                        temporal_model.value, "*", str(deg_id),
                        "s-*-single.jsonl"))

  print(os.path.join(
      		    input_prefix, static_model.value,
      		    temporal_model.value, "*", str(deg_id),
      		    "s-*-single.jsonl"))

  for filename in tqdm.tqdm(list(single_files)):
    with open(filename) as f:
      pattern = os.path.join(input_prefix, static_model.value,
                               temporal_model.value, r"(\d+)",
                               str(deg_id), r".*\.jsonl")
      m = re.fullmatch(pattern, filename)
      size = int(m[1])  # if this fails then the pattern is wrong :/
      for line in f:
        try:
          s = json.loads(line)
          if all(k in s for k in single_required) and \
            all(k in s["network"] for k in network_required):
            try:
              prob_dist = s["prob_dist_type"]
              dt = s["dt"]
              if size not in single_summaries:
                single_summaries[size] = {}
              if prob_dist not in single_summaries[size]:
                single_summaries[size][prob_dist] = {}
              if dt not in single_summaries[size][prob_dist]:
                single_summaries[size][prob_dist][dt] = {k: welfords() for k in single_required}
                single_summaries[size][prob_dist][dt]["network"] = {k: welfords() for k in network_required}

              for k in single_required:
                if k == "network":
                  for j in network_required:
                    single_summaries[size][prob_dist][dt][k][j].update(s[k][j])
                else:
                  single_summaries[size][prob_dist][dt][k].update(s[k])
            except ValueError:
              continue
          else:
            print("missing key:", [k for k in single_required if k not in s])
        except json.JSONDecodeError:
          continue

  print("sizes: ", sorted(single_summaries.keys()))

  for size in single_summaries:
    for prob_dist in single_summaries[size]:
      for dt in single_summaries[size][prob_dist]:
        summaries = single_summaries[size][prob_dist][dt]
        agg = {}
        for k in single_required:
          if k == "network":
            agg["network"] = {}
            for j in network_required:
              mean = summaries[k][j].mean()
              sem = summaries[k][j].sem()
              agg[k][j] = (mean, sem)
          else:
            mean = summaries[k].mean()
            sem = summaries[k].sem()
            agg[k] = (mean, sem)

        # make sure there are no duplicate seeds present
        # print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
        # assert len(summaries["seed"]) == len(set(summaries["seed"]))

        # calculate Kaplan–Meier survival function
        independent_samples = summaries["seed"].count()
        km_survival_mean = []
        km_survival_sem = []
        lifetime_dist = summaries["lifetime-dist"].sum()
        censored_lifetime_dist = summaries["censored-lifetime-dist"].sum()
        uncensored_lifetime_dist = lifetime_dist - censored_lifetime_dist
        last_sum = 0.0
        last_survival = 1.0
        for i in range(len(agg["time-bin-edges"][0]) - 1):
          n_i = np.sum(lifetime_dist[i:])
          d_i = uncensored_lifetime_dist[i]
          if n_i > 0 and n_i > d_i:
            last_survival = last_survival*(1.0 - d_i/n_i)
            last_sum = last_sum + (d_i/n_i)*(d_i/(n_i - d_i))
            km_survival_mean.append(last_survival)
            km_survival_sem.append(
                last_survival*np.sqrt(last_sum)/np.sqrt(independent_samples))
          else:
            km_survival_mean.append(np.nan)
            km_survival_sem.append(np.nan)
        agg["km-survival-dist"] = (np.array(km_survival_mean), np.array(km_survival_sem))
        
        single_summaries[size][prob_dist][dt] = agg

  output_path = os.path.join(output_prefix,
      f"{static_model.value}-{temporal_model.value}-{deg_id}-single.json")
  with open(output_path, 'w') as f:
    json.dump(single_summaries, f, cls=NumpyEncoder)

  # homogeneous_summaries = {}
  # homogeneous_files = glob.glob(os.path.join(
  #                       input_prefix, static_model,
  #                       temporal_model, "*", str(deg_id),
  #                       f"s-*-homogeneous.jsonl"))

  # for filename in tqdm.tqdm(list(homogeneous_files)):
  #   with open(filename) as f:
  #     pattern = os.path.join(input_prefix, static_model,
  #                              temporal_model, r"(\d+)",
  #                              str(deg_id), r".*\.jsonl")
  #     m = re.fullmatch(pattern, filename)
  #     size = int(m[1])  # if this fails then the pattern is wrong :/
  #     for line in f:
  #       try:
  #         s = json.loads(line)
  #         if all(k in s for k in homogeneous_required) and \
  #           all(k in s["network"] for k in network_required):
  #           try:
  #             prob_dist = s["prob_dist_type"]
  #             dt = s["dt"]
  #             field = s["field"]
  #             if size not in homogeneous_summaries:
  #               homogeneous_summaries[size] = {}
  #             if prob_dist not in homogeneous_summaries[size]:
  #               homogeneous_summaries[size][prob_dist] = {}
  #             if dt not in homogeneous_summaries[size][prob_dist]:
  #               homogeneous_summaries[size][prob_dist][dt] = {}
  #             if field not in homogeneous_summaries[size][prob_dist][dt]:
  #               homogeneous_summaries[size][prob_dist][dt][field] = {k: [] for k in homogeneous_required}
  #               homogeneous_summaries[size][prob_dist][dt][field]["network"] = {k: [] for k in network_required}

  #             for k in homogeneous_required:
  #               if k == "network":
  #                 for j in network_required:
  #                   homogeneous_summaries[size][prob_dist][dt][field][k][j].append(s[k][j])
  #               else:
  #                 homogeneous_summaries[size][prob_dist][dt][field][k].append(s[k])
  #           except ValueError:
  #             continue
  #         else:
  #           print("missing key:", [k for k in homogeneous_required if k not in s])
  #       except json.JSONDecodeError:
  #         continue

  # for size in homogeneous_summaries:
  #   for prob_dist in homogeneous_summaries[size]:
  #     for dt in homogeneous_summaries[size][prob_dist]:
  #       for field in homogeneous_summaries[size][prob_dist][dt]:
  #         summaries = homogeneous_summaries[size][prob_dist][dt][field]

  #         agg = {}
  #         for k in homogeneous_required:
  #           if k == "network":
  #             agg["network"] = {}
  #             for j in network_required:
  #               mean = np.mean(summaries[k][j], axis=0)
  #               sem = stats.sem(summaries[k][j], axis=0)
  #               agg[k][j] = (mean, sem)
  #           else:
  #             mean = np.mean(to_array(summaries[k]), axis=0)
  #             sem = stats.sem(to_array(summaries[k]), axis=0)
  #             agg[k] = (mean, sem)

  #         # make sure all have the same time-bins
  #         assert all(np.allclose(edges, agg["time-bin-edges"][0])
  #                      for edges in summaries["time-bin-edges"])

  #         # make sure there are no duplicate seeds present
  #         print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
  #         # assert len(summaries["seed"]) == len(set(summaries["seed"]))

  #         bins = agg["time-bin-edges"][0]
  #         threshold_bins = 16  # use first (last) n bins to calculate density
  #         last_bin_hypervolume = size*(bins[threshold_bins+1]-bins[0])

  #         stat_densities = np.sum(
  #           summaries["occupied-event-times-to-end-dist"][:threshold_bins],
  #           axis=1)/last_bin_hypervolume
  #         agg["static_density"] = (np.mean(stat_densities), stats.sem(stat_densities))

  #         homogeneous_summaries[size][prob_dist][dt][field] = agg

  # output_path = os.path.join(output_prefix,
  #     f"{static_model}-{temporal_model}-{deg_id}-homogeneous.json")
  # with open(output_path, 'w') as f:
  #   json.dump(homogeneous_summaries, f, cls=NumpyEncoder)


  homogeneous_alt_summaries = {}
  homogeneous_alt_files = glob.glob(os.path.join(
                        input_prefix, static_model.value,
                        temporal_model.value, "*", str(deg_id),
                        f"s-*-homogeneous.jsonl"))

  for filename in tqdm.tqdm(list(homogeneous_alt_files)):
    with open(filename) as f:
      pattern = os.path.join(input_prefix, static_model.value,
                               temporal_model.value, r"(\d+)",
                               str(deg_id), r".*\.jsonl")
      m = re.fullmatch(pattern, filename)
      size = int(m[1])  # if this fails then the pattern is wrong :/
      for line in f:
        try:
          s = json.loads(line)
          if all(k in s for k in homogeneous_alt_required) and \
            all(k in s["network"] for k in network_required):
            try:
              prob_dist = s["prob_dist_type"]
              dt = s["dt"]
              if size not in homogeneous_alt_summaries:
                homogeneous_alt_summaries[size] = {}
              if prob_dist not in homogeneous_alt_summaries[size]:
                homogeneous_alt_summaries[size][prob_dist] = {}
              if dt not in homogeneous_alt_summaries[size][prob_dist]:
                homogeneous_alt_summaries[size][prob_dist][dt] = {k: welfords() for k in homogeneous_alt_required}
                homogeneous_alt_summaries[size][prob_dist][dt]["network"] = {k: welfords() for k in network_required}

              for k in homogeneous_alt_required:
                if k == "network":
                  for j in network_required:
                    homogeneous_alt_summaries[size][prob_dist][dt][k][j].update(s[k][j])
                else:
                  homogeneous_alt_summaries[size][prob_dist][dt][k].update(s[k])
            except ValueError:
              continue
          else:
            print("missing key:", [k for k in homogeneous_alt_required if k not in s])
        except json.JSONDecodeError:
          continue

  print("sizes: ", sorted(homogeneous_alt_summaries.keys()))

  for size in homogeneous_alt_summaries:
    for prob_dist in homogeneous_alt_summaries[size]:
      for dt in homogeneous_alt_summaries[size][prob_dist]:
        summaries = homogeneous_alt_summaries[size][prob_dist][dt]

        agg = {}
        for k in homogeneous_alt_required:
          if k == "network":
            agg["network"] = {}
            for j in network_required:
              mean = summaries[k][j].mean()
              sem = summaries[k][j].sem()
              agg[k][j] = (mean, sem)
          else:
            mean = summaries[k].mean()
            sem = summaries[k].sem()
            agg[k] = (mean, sem)

        # make sure there are no duplicate seeds present
        # print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
        # assert len(summaries["seed"]) == len(set(summaries["seed"]))

        bins = agg["time-bin-edges"][0]
        threshold_bins = 16  # use first (last) n bins to calculate density
        last_bin_hypervolume = size*(bins[threshold_bins+1]-bins[0])
        dt_band_hypervolume = size*agg["dt"][0]

        dt_band_stat_densities = agg["dt-band-homogeneous-occupied"][0]/dt_band_hypervolume
        agg["dt-band-static-density"] = (np.mean(dt_band_stat_densities), stats.sem(dt_band_stat_densities))

        dt_band_susceptibility = agg["network"]["temporal-edges"][0] * (
          agg["dt-band-external-field-occupied"][0] - agg["dt-band-homogeneous-occupied"][0])/dt_band_hypervolume
        agg["dt-band-susceptibility"] = (np.mean(dt_band_susceptibility), stats.sem(dt_band_susceptibility))

        hypervolumes = (bins[1:] - bins[:-1])*size
        densities = summaries["homogeneous-occupied-count-dist"]
        agg["density"] = (densities.mean()/hypervolumes, densities.sem()/np.abs(hypervolumes))

        homogeneous_alt_summaries[size][prob_dist][dt] = agg

  output_path = os.path.join(output_prefix,
      f"{static_model.value}-{temporal_model.value}-{deg_id}-homogeneous.json")
  with open(output_path, 'w') as f:
    json.dump(homogeneous_alt_summaries, f, cls=NumpyEncoder)

  dist_summaries = {}
  dist_files = glob.glob(os.path.join(
                        input_prefix, static_model.value,
                        temporal_model.value, "*", str(deg_id),
                        "s-*-degree-dist.jsonl"))

  print(os.path.join(
      		    input_prefix, static_model.value,
      		    temporal_model.value, "*", str(deg_id),
      		    "s-*-degree-dist.jsonl"))

  for filename in tqdm.tqdm(list(dist_files)):
    with open(filename) as f:
      pattern = os.path.join(input_prefix, static_model.value,
                               temporal_model.value, r"(\d+)",
                               str(deg_id), r".*\.jsonl")
      m = re.fullmatch(pattern, filename)
      size = int(m[1])  # if this fails then the pattern is wrong :/
      for line in f:
        try:
          s = json.loads(line)
          if all(k in s for k in dist_required) and \
            all(k in s["network"] for k in network_required):
            try:
              prob_dist = s["prob_dist_type"]
              dt = s["dt"]
              if size not in dist_summaries:
                dist_summaries[size] = {}
              if prob_dist not in dist_summaries[size]:
                dist_summaries[size][prob_dist] = {}
              if dt not in dist_summaries[size][prob_dist]:
                dist_summaries[size][prob_dist][dt] = {k: welfords() for k in dist_required}
                dist_summaries[size][prob_dist][dt]["network"] = {k: welfords() for k in network_required}
                dist_summaries[size][prob_dist][dt]["counts"] = {}

              for k in dist_required:
                if k == "network":
                  for j in network_required:
                    dist_summaries[size][prob_dist][dt][k][j].update(s[k][j])
                elif k == "counts":
                  for cnt in s[k]:
                    if (cnt["in"], cnt["out"]) not in dist_summaries[size][prob_dist][dt]["counts"]:
                      dist_summaries[size][prob_dist][dt]["counts"][cnt["in"], cnt["out"]] = 0
                    dist_summaries[size][prob_dist][dt]["counts"][(cnt["in"], cnt["out"])] += cnt["count"]
                else:
                  dist_summaries[size][prob_dist][dt][k].update(s[k])
            except ValueError:
              continue
          else:
            print("missing key:", [k for k in dist_required if k not in s])
        except json.JSONDecodeError:
          continue

  print("sizes: ", sorted(dist_summaries.keys()))

  for size in dist_summaries:
    for prob_dist in dist_summaries[size]:
      for dt in dist_summaries[size][prob_dist]:
        summaries = dist_summaries[size][prob_dist][dt]
        agg = {}
        for k in dist_required:
          if k == "network":
            agg["network"] = {}
            for j in network_required:
              mean = summaries[k][j].mean()
              sem = summaries[k][j].sem()
              agg[k][j] = (mean, sem)
          elif k == "counts":
            agg[k] = []
            for (i, j), c in summaries["counts"].items():
              agg[k].append({"in": i, "out": j, "count": c}) 
          else:
            mean = summaries[k].mean()
            sem = summaries[k].sem()
            agg[k] = (mean, sem)

        # make sure there are no duplicate seeds present
        # print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
        # assert len(summaries["seed"]) == len(set(summaries["seed"]))
        dist_summaries[size][prob_dist][dt] = agg

  output_path = os.path.join(output_prefix,
      f"{static_model.value}-{temporal_model.value}-{deg_id}-degree-dist.json")
  with open(output_path, 'w') as f:
    json.dump(dist_summaries, f, cls=NumpyEncoder)



if __name__ == "__main__":
  parser = argparse.ArgumentParser(
      description='aggreagates separate results into fewer, bigger files')
  parser.add_argument('input_prefix', type=str)
  parser.add_argument('output_prefix', type=str)

  args = parser.parse_args()

  sources = chain(
	# product(
        #   [StaticModel.grid_1],
        #   [TemporalModel.poisson],
        #   [2]),
	# product(
        #   [StaticModel.grid_2],
        #   [TemporalModel.poisson],
        #   [4]),
	# product(
        #   [StaticModel.grid_3],
        #   [TemporalModel.poisson],
        #   [6]),
	# product(
        #   [StaticModel.grid_4],
        #   [TemporalModel.poisson],
        #   [8]),
	# product(
        #   StaticModel.heavytails(),
        #   [TemporalModel.poisson],
        #   [9]),
	# product(
        #   [StaticModel.regular],
        #   [# TemporalModel.poisson,
	#     # *TemporalModel.bursties(),
	#     *TemporalModel.self_excitings()
	#  ],
        #  [9]),
	 product(
           [StaticModel.erdos],
           [TemporalModel.poisson,
	    #  *TemporalModel.bursties(),
	    #  *TemporalModel.self_excitings()
           ],
           [8]),
	# product(
        #   [StaticModel.nsi_regular],
        #   [TemporalModel.poisson,
	#     # *TemporalModel.bursties(),
	#     # *TemporalModel.self_excitings()
	#  ],
        #  [9]),
        )


  os.makedirs(args.output_prefix, exist_ok=True)

  with Pool() as pool:
    pool.starmap(load_data, zip(repeat(args.input_prefix), repeat(args.output_prefix), *zip(*sources)))
