import os
import re
import glob
import argparse
import json
import collections
from itertools import product, chain
import enum

import numpy as np
from scipy import stats
import tqdm

def to_array(num_array):
  return np.array(num_array, dtype=np.float)

class NumpyEncoder(json.JSONEncoder):
  def default(self, obj):
   if type(obj) == float and np.isnan(val):
     return None
   if isinstance(obj, np.ndarray):
     return obj.tolist()
   return json.JSONEncoder.default(self, obj)

def load_data(input_prefix, output_prefix, size):


  os.makedirs(output_prefix, exist_ok=True)

  single_required = [
      "dt", "seed", "window",
      "network", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "lifetime-dist",
      "mean-mass", "mean-volume", "mean-lifetime", "mass-times-dist", "volume-times-dist",
      "mass-times-to-end-dist", "volume-times-to-end-dist",
      "censored-event-times-dist", "censored-lifetime-dist"]

  homogeneous_required = [
      "dt", "seed", "window",
      "network", "field", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "occupied-event-times-dist",
      "occupied-event-times-to-end-dist"]

  homogeneous_alt_required = [
      "dt", "seed", "window",
      "network", "time-bin-edges",
      "event-times-dist", "event-times-to-end-dist",
      "dt-band-homogeneous-occupied", "dt-band-external-field-occupied",
      "homogeneous-occupied-count-dist", "homogeneous-occupied-count-to-end-dist",
      "external-field-occupied-count-dist", "external-field-occupied-count-to-end-dist"]

  network_required = ["temporal-edges", "temporal-nodes"]

  single_summaries = {}
  single_files = glob.glob(os.path.join(
                        input_prefix,
                        "single.jsonl"))

  for filename in tqdm.tqdm(list(single_files)):
    with open(filename) as f:
      for line in f:
        try:
          s = json.loads(line)
          if all(k in s for k in single_required) and \
            all(k in s["network"] for k in network_required):
            try:
              prob_dist = s["prob_dist_type"]
              dt = s["dt"]
              if size not in single_summaries:
                single_summaries[size] = {}
              if prob_dist not in single_summaries[size]:
                single_summaries[size][prob_dist] = {}
              if dt not in single_summaries[size][prob_dist]:
                single_summaries[size][prob_dist][dt] = {k: [] for k in single_required}
                single_summaries[size][prob_dist][dt]["network"] = {k: [] for k in network_required}

              for k in single_required:
                if k == "network":
                  for j in network_required:
                    single_summaries[size][prob_dist][dt][k][j].append(s[k][j])
                else:
                  single_summaries[size][prob_dist][dt][k].append(s[k])
            except ValueError:
              continue
          else:
            print("missing key:", [k for k in single_required if k not in s])
        except json.JSONDecodeError:
          continue

  print("sizes: ", single_summaries.keys())

  for size in single_summaries:
    for prob_dist in single_summaries[size]:
      for dt in single_summaries[size][prob_dist]:
        summaries = single_summaries[size][prob_dist][dt]
        agg = {}
        for k in single_required:
          if k == "network":
            agg["network"] = {}
            for j in network_required:
              mean = np.mean(summaries[k][j], axis=0)
              sem = stats.sem(summaries[k][j], axis=0)
              median = np.median(summaries[k][j], axis=0)
              agg[k][j] = (mean, sem, median)
          else:
            mean = np.mean(to_array(summaries[k]), axis=0)
            sem = stats.sem(to_array(summaries[k]), axis=0)
            median = np.median(to_array(summaries[k]), axis=0)
            agg[k] = (mean, sem, median)

        # make sure all have the same time-bins
        assert all(np.allclose(edges, agg["time-bin-edges"][0])
                     for edges in summaries["time-bin-edges"])

        # make sure there are no duplicate seeds present
        # print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
        # assert len(summaries["seed"]) == len(set(summaries["seed"]))

        # calculate Kaplan–Meier survival function
        independent_samples = len(summaries["seed"])
        km_survival_mean = []
        km_survival_sem = []
        lifetime_dist = np.sum(summaries["lifetime-dist"], axis=0)
        censored_lifetime_dist = np.sum(summaries["censored-lifetime-dist"], axis=0)
        uncensored_lifetime_dist = lifetime_dist - censored_lifetime_dist
        last_sum = 0.0
        last_survival = 1.0
        for i in range(len(agg["time-bin-edges"][0]) - 1):
          n_i = np.sum(lifetime_dist[i:])
          d_i = uncensored_lifetime_dist[i]
          if n_i > 0 and n_i > d_i:
            last_survival = last_survival*(1.0 - d_i/n_i)
            last_sum = last_sum + (d_i/n_i)*(d_i/(n_i - d_i))
            km_survival_mean.append(last_survival)
            km_survival_sem.append(
                last_survival*np.sqrt(last_sum)/np.sqrt(independent_samples))
          else:
            km_survival_mean.append(np.nan)
            km_survival_sem.append(np.nan)
        agg["km-survival-dist"] = (np.array(km_survival_mean), np.array(km_survival_sem))
        
        single_summaries[size][prob_dist][dt] = agg

  output_path = os.path.join(output_prefix,
      f"single.json")
  with open(output_path, 'w') as f:
    json.dump(single_summaries, f, cls=NumpyEncoder)


  homogeneous_alt_summaries = {}
  homogeneous_alt_files = glob.glob(os.path.join(
                        input_prefix,
                        "homogeneous-alt.jsonl"))

  for filename in tqdm.tqdm(list(homogeneous_alt_files)):
    with open(filename) as f:
      for line in f:
        try:
          s = json.loads(line)
          if all(k in s for k in homogeneous_alt_required) and \
            all(k in s["network"] for k in network_required):
            try:
              prob_dist = s["prob_dist_type"]
              dt = s["dt"]
              if size not in homogeneous_alt_summaries:
                homogeneous_alt_summaries[size] = {}
              if prob_dist not in homogeneous_alt_summaries[size]:
                homogeneous_alt_summaries[size][prob_dist] = {}
              if dt not in homogeneous_alt_summaries[size][prob_dist]:
                homogeneous_alt_summaries[size][prob_dist][dt] = {k: [] for k in homogeneous_alt_required}
                homogeneous_alt_summaries[size][prob_dist][dt]["network"] = {k: [] for k in network_required}

              for k in homogeneous_alt_required:
                if k == "network":
                  for j in network_required:
                    homogeneous_alt_summaries[size][prob_dist][dt][k][j].append(s[k][j])
                else:
                  homogeneous_alt_summaries[size][prob_dist][dt][k].append(s[k])
            except ValueError:
              continue
          else:
            print("missing key:", [k for k in homogeneous_alt_required if k not in s])
        except json.JSONDecodeError:
          continue

  print("sizes: ", homogeneous_alt_summaries.keys())

  for size in homogeneous_alt_summaries:
    for prob_dist in homogeneous_alt_summaries[size]:
      for dt in homogeneous_alt_summaries[size][prob_dist]:
        summaries = homogeneous_alt_summaries[size][prob_dist][dt]

        agg = {}
        for k in homogeneous_alt_required:
          if k == "network":
            agg["network"] = {}
            for j in network_required:
              mean = np.mean(summaries[k][j], axis=0)
              sem = stats.sem(summaries[k][j], axis=0)
              median = np.median(summaries[k][j], axis=0)
              agg[k][j] = (mean, sem, median)
          else:
            mean = np.mean(to_array(summaries[k]), axis=0)
            sem = stats.sem(to_array(summaries[k]), axis=0)
            median = np.median(to_array(summaries[k]), axis=0)
            agg[k] = (mean, sem, median)

        # make sure all have the same time-bins
        assert all(np.allclose(edges, agg["time-bin-edges"][0])
                     for edges in summaries["time-bin-edges"])

        # make sure there are no duplicate seeds present
        # print("dupes for size ", size, [item for item, count in collections.Counter(summaries["seed"]).items() if count > 1])
        # assert len(summaries["seed"]) == len(set(summaries["seed"]))

        bins = agg["time-bin-edges"][0]
        threshold_bins = 1  # use first (last) n bins to calculate density
        last_bin_hypervolume = size*(bins[threshold_bins+1]-bins[0])
        dt_band_hypervolume = size*agg["dt"][0]

        stat_densities = np.sum(
          summaries["homogeneous-occupied-count-to-end-dist"][:threshold_bins],
          axis=1)/last_bin_hypervolume
        agg["static-density"] = (np.mean(stat_densities), stats.sem(stat_densities), np.median(stat_densities))

        dt_band_stat_densities = agg["dt-band-homogeneous-occupied"][0]/dt_band_hypervolume
        agg["dt-band-static-density"] = (np.mean(dt_band_stat_densities), stats.sem(dt_band_stat_densities), np.median(dt_band_stat_densities))

        susceptibility = agg["network"]["temporal-edges"][0] * (np.sum(
          summaries["external-field-occupied-count-to-end-dist"][:threshold_bins],
          axis=1)/last_bin_hypervolume - stat_densities)
        agg["susceptibility"] = (np.mean(susceptibility), stats.sem(susceptibility), np.median(susceptibility))

        dt_band_susceptibility = agg["network"]["temporal-edges"][0] * (
          agg["dt-band-external-field-occupied"][0] - agg["dt-band-homogeneous-occupied"][0])/dt_band_hypervolume
        agg["dt-band-susceptibility"] = (np.mean(dt_band_susceptibility), stats.sem(dt_band_susceptibility), np.median(dt_band_susceptibility))

        hypervolumes = (bins[1:] - bins[:-1])*size
        densities = np.array(summaries["homogeneous-occupied-count-dist"])/hypervolumes
        agg["density"] = (np.mean(densities, axis=0), stats.sem(densities, axis=0), np.median(densities, axis=0))

        homogeneous_alt_summaries[size][prob_dist][dt] = agg

  output_path = os.path.join(output_prefix,
      f"homogeneous-alt.json")
  with open(output_path, 'w') as f:
    json.dump(homogeneous_alt_summaries, f, cls=NumpyEncoder)


class RealNetworks(enum.Enum):
  air = "air"
  transport = "transport"
  twitter = "twitter"
  mobile = "mobile"

if __name__ == "__main__":
  parser = argparse.ArgumentParser(
      description='aggreagates separate results into fewer, bigger files')
  parser.add_argument('network', type=RealNetworks)

  args = parser.parse_args()

  input_prefix = os.path.join("networks", args.network.value)
  output_prefix = input_prefix

  size = 1
  if args.network is RealNetworks.air:
    size = 279
  elif args.network is RealNetworks.transport:
    size = 6986
  elif args.network is RealNetworks.twitter:
    size = 17313623
  elif args.network is RealNetworks.mobile:
    size = 5193086
  load_data(input_prefix, output_prefix, size)
