import json
import argparse
from collections import defaultdict
import sys
import csv

from scipy import interpolate
from scipy import stats
import numpy as np
import tqdm
import matplotlib.pyplot as plt

from network_configs import StaticModel, TemporalModel, ProbDist 

def all_dts(single_summaries, homogeneous_summaries, prob_dist):
  dts = []
  for name, summaries in [("single", single_summaries), ("homogeneous", homogeneous_summaries)]:
    sizes = sorted(summaries, key=int)
    for size_s in sizes:
      s = summaries[size_s][prob_dist.value]
      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts.append(set(keys))
      print(f"{name} {size_s} dtcs:", keys, file=sys.stderr)

  return set.intersection(*dts)

def fit_exponent(single_summaries, homogeneous_summaries, prob_dist, dtcs, nu_para, nu_perp, dims, survival_ax, density_ax):
  max_size_s = max(single_summaries, key=int)
  size = int(max_size_s)
  dtc = dtcs[size]
  s = single_summaries[max_size_s][prob_dist.value]
  time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])
  mk_survival = np.array(s[str(dtc)]["km-survival-dist"][0])
  max_t = 0.04*size**(nu_para[dims]/(nu_perp[dims]*dims))
  min_t = 2e0
  mask = (time > min_t) & (time < max_t)
  res = stats.linregress(np.log(time[mask]), np.log(mk_survival[mask]))
  delta = -res.slope
  survival_ax.plot(time, mk_survival, ls='--', marker='.')
  survival_ax.plot(time, np.exp(res.intercept)*time**(-delta), ls='-', marker='')
  survival_ax.set_xscale("log")
  survival_ax.set_yscale("log")
  survival_ax.set_xlabel("Time")
  survival_ax.set_ylabel("$\\hat{P}(t)$")

  max_size_s = max(homogeneous_summaries, key=int)
  size = int(max_size_s)
  dtc = dtcs[size]
  s = homogeneous_summaries[max_size_s][prob_dist.value]
  time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])
  density = np.array(s[str(dtc)]["density"][0])
  max_t = 0.04*size**(nu_para[dims]/(nu_perp[dims]*dims))
  min_t = 2e0
  mask = (time > min_t) & (time < max_t)
  res = stats.linregress(np.log(time[mask]), np.log(density[mask]))
  alpha = -res.slope
  density_ax.plot(time, density, ls='--', marker='.')
  density_ax.plot(time, np.exp(res.intercept)*time**(-alpha), ls='-', marker='')
  density_ax.set_xscale("log")
  density_ax.set_yscale("log")
  density_ax.set_xlabel("Time")
  density_ax.set_ylabel("$\\rho(t)$")
    
  return alpha, delta
    

def fit_collapse(single_summaries, homogeneous_summaries, prob_dist, dtcs, beta, beta_p, nu_para, nu_perp, dims, max_size=np.inf):
  delta = beta_p[dims]/nu_para[dims]
  theta = dims*nu_para[dims] - beta[dims] - beta_p[dims]
  gamma = nu_para[dims] + dims*nu_perp[dims] - beta[dims] - beta_p[dims]
  upsilon = dims*nu_perp[dims] - beta_p[dims]

  sizes = sorted(single_summaries, key=int)
  sizes = [s for s in sizes if int(s) <= max_size]

  times = {}
  mk_survivals = {}
  mk_survival_interp = {}

  masses = {}
  mass_interp = {}

  volumes = {}
  volume_interp = {}

  time_min = -np.inf
  time_max = np.inf

  for size_s in sizes:
    s = single_summaries[size_s][prob_dist.value]
    size = int(size_s)

    keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
    dts = np.array([s[k]["dt"][0] for k in keys])
    dtc = dtcs[size]
    taus = (dts - dtc)

    mk_survival_mask = (np.array(s[str(dtc)]["km-survival-dist"][0]) > 0)
    mass_mask = (np.array(s[str(dtc)]["mass-times-to-end-dist"][0]) > 0)
    volume_mask = (np.array(s[str(dtc)]["volume-times-to-end-dist"][0]) > 0)
    mask = (mk_survival_mask & mass_mask & volume_mask)

    scaled_time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])*size**(-nu_para[dims]/(nu_perp[dims]*dims))
    scaled_mk_survival = np.array(s[str(dtc)]["km-survival-dist"][0])*size**(beta_p[dims]/(nu_perp[dims]*dims))
    scaled_mass = np.array(s[str(dtc)]["mass-times-to-end-dist"][0])*size**(-gamma/(nu_perp[dims]*dims))
    scaled_volume = np.array(s[str(dtc)]["volume-times-to-end-dist"][0])*size**(-upsilon/(nu_perp[dims]*dims))

    time_min = max(time_min, np.amin(scaled_time[mask]))
    time_max = min(time_max, np.amax(scaled_time[mask]))
    mk_survival_interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_mk_survival[mask]))
    mass_interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_mass[mask]))
    volume_interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_volume[mask]))
    times[size_s] = scaled_time[mask]
    mk_survivals[size_s] = scaled_mk_survival[mask]
    masses[size_s] = scaled_mass[mask]
    volumes[size_s] = scaled_volume[mask]

  mean_survival = lambda t: np.mean([f(t) for f in mk_survival_interp.values()])
  mean_mass = lambda t: np.mean([f(t) for f in mass_interp.values()])
  mean_volume = lambda t: np.mean([f(t) for f in volume_interp.values()])

  survival_ssd = 0.0
  mass_ssd = 0.0
  volume_ssd = 0.0

  count = 0
  for s in sizes:
    for t, p, m, v in zip(times[s], mk_survivals[s], masses[s], volumes[s]):
      if t >= time_min and t <= time_max:
        survival_ssd += (mean_survival(np.log(t)) - np.log(p))**2
        mass_ssd += (mean_mass(np.log(t)) - np.log(m))**2
        volume_ssd += (mean_volume(np.log(t)) - np.log(v))**2
        count += 1

  survival_msd = 0.0
  mass_msd = 0.0
  volume_msd = 0.0
  if count > 0:
    survival_msd = survival_ssd/count
    mass_msd = mass_ssd/count
    volume_msd = volume_ssd/count
  else:
    survival_msd = np.inf
    mass_ssd = np.inf
    volume_msd = np.inf

  sizes = sorted(homogeneous_summaries, key=int)
  sizes = [s for s in sizes if int(s) <= max_size]

  interp = {}
  times = {}
  densities = {}

  time_min = -np.inf
  time_max = np.inf

  for size_s in sizes:
    s = homogeneous_summaries[size_s][prob_dist.value]
    size = int(size_s)

    keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
    dts = np.array([s[k]["dt"][0] for k in keys])
    dtc = dtcs[size]
    taus = (dts - dtc)

    mask = (np.array(s[str(dtc)]["density"][0]) > 0)

    scaled_time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])*size**(-nu_para[dims]/(nu_perp[dims]*dims))
    scaled_density = np.array(s[str(dtc)]["density"][0])*size**(beta[dims]/(nu_perp[dims]*dims))

    if len(scaled_time[mask]) == 0:
      continue

    time_min = max(time_min, np.amin(scaled_time[mask]))
    time_max = min(time_max, np.amax(scaled_time[mask]))
    interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_density[mask]))
    times[size_s] = scaled_time[mask]
    densities[size_s] = scaled_density[mask]

  mean_density = lambda t: np.mean([interp[s](t) for s in interp])
  count = 0
  ssd = 0.0
  for s in sizes:
    if s in times and s in densities:
      for t, d in zip(times[s], densities[s]):
        if t >= time_min and t <= time_max:
          ssd += (mean_density(np.log(t)) - np.log(d))**2
          count += 1

  density_msd = 0.0
  if count > 0:
    density_msd = ssd/count
  else:
    density_msd = np.inf

  return np.sqrt(survival_msd), np.sqrt(density_msd), np.sqrt(mass_msd), np.sqrt(volume_msd)

if __name__=="__main__":
  parser = argparse.ArgumentParser(description="Plots quantities of random networks")
  parser.add_argument('static', type=StaticModel)
  parser.add_argument('temporal', type=TemporalModel)
  args = parser.parse_args()

  beta =    {1: 0.276, 2: 0.583, 3: 0.813, 4: 1.0}
  nu_perp = {1: 1.096, 2: 0.733, 3: 0.584, 4: 0.5}
  nu_para = {1: 1.734, 2: 1.295, 3: 1.11,  4: 1.0}
  
  static = args.static
  temporal = args.temporal
  prob_dist = ProbDist.deterministic
  deg_id = 9
  min_dt = 0.08
  max_dt = 0.1
  dims = 4
  max_collapse_size = np.inf

  if static is StaticModel.nsi_regular:
    min_dt = 0.0
    max_dt = 0.9

  if static in [StaticModel.regular, StaticModel.erdos] and temporal is TemporalModel.bursty_205:
    min_dt = 0.05
    max_dt = 0.08

  if static in [StaticModel.regular, StaticModel.erdos] and temporal is TemporalModel.bursty_220:
    min_dt = 0.07
    max_dt = 0.1

  if static in [StaticModel.regular, StaticModel.erdos] and \
	  temporal in [TemporalModel.self_exciting_02_08_10, TemporalModel.self_exciting_02_08_05]:
    min_dt = 0.07
    max_dt = 0.1

  if static is StaticModel.erdos:
    deg_id = 8

  if static is StaticModel.grid_1:
    min_dt = 0.0
    max_dt = 2.0
    dims = 1
    deg_id = 2
    max_collapse_size = 2**12

  if static is StaticModel.grid_2:
    min_dt = 0.25
    max_dt = 0.35
    dims = 2
    deg_id = 4
    max_collapse_size = 2**12

  if static is StaticModel.grid_3:
    min_dt = 0.13
    max_dt = 0.19
    dims = 3
    deg_id = 6
    max_collapse_size = 2**13

  if static is StaticModel.grid_4:
    min_dt = 0.07
    max_dt = 0.12
    dims = 4
    deg_id = 8
    max_collapse_size = 2**12

  single_filename = f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-single.json"
  homogeneous_filename = f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-homogeneous.json"

  with open(single_filename) as sngl_f, open(homogeneous_filename) as homo_f:
    single_summaries = json.load(sngl_f)
    homogeneous_summaries = json.load(homo_f)

  
    dtc_candidates = all_dts(single_summaries, homogeneous_summaries, prob_dist)
    dtc_candidates = [c for c in sorted(dtc_candidates) if float(c) >= min_dt and float(c) <= max_dt]
    print("dtcs:", dtc_candidates, file=sys.stderr)
    
    survival_ssd = []
    density_ssd = []
    mass_ssd = []
    volume_ssd = []
    for k in tqdm.tqdm(dtc_candidates):
      s1, s2, s3, s4 = fit_collapse(single_summaries, homogeneous_summaries, prob_dist, dtcs=defaultdict(lambda: float(k)),
          beta=beta, beta_p=beta, nu_perp=nu_perp, nu_para=nu_para, dims=dims, max_size=max_collapse_size)
      survival_ssd.append(s1)
      density_ssd.append(s2)
      mass_ssd.append(s3)
      volume_ssd.append(s4)

    opt_dt, min_s1, min_s2, min_s3, min_s4 = min(zip(dtc_candidates, survival_ssd, density_ssd, mass_ssd, volume_ssd), key=lambda a: a[1] + a[2] + a[3] + a[4])

    
    fig, ax = plt.subplots(figsize=(4.2, 3.4))
    ax.axvline(float(opt_dt), ls='--', label=f"$\\delta t = {opt_dt}$", color='gray')
    ax.plot(list(map(float, dtc_candidates)), survival_ssd, label="$RMSE[\\hat{P}(t)]$", marker='.')
    ax.plot(list(map(float, dtc_candidates)), density_ssd, label="$RMSE[\\rho(t)]$", marker='.')
    ax.plot(list(map(float, dtc_candidates)), mass_ssd, label="$RMSE[M(t)]$", marker='.')
    ax.plot(list(map(float, dtc_candidates)), volume_ssd, label="$RMSE[V(t)]$", marker='.')
    delta = 4e-4
    delta = 6e-3
    ax.set_xlim(float(opt_dt)-delta, float(opt_dt)+delta)
    ax.set_xlabel("$\\delta t$")
    ax.set_ylabel("RMS Error w/ mean curve")
    ax.set_yscale("log")

    ax.legend()
    fig.savefig(f"figures/fit-{static.value}-{temporal.value}-{deg_id}.pdf",
        bbox_inches='tight', pad_inches=0.1)

    fig, (survival_ax, density_ax) = plt.subplots(nrows=2, figsize=(4.2, 7.0))
    alpha, delta = fit_exponent(single_summaries, homogeneous_summaries, prob_dist, dtcs=defaultdict(lambda: float(opt_dt)),
        nu_para=nu_para, nu_perp=nu_perp, dims=dims, density_ax=density_ax, survival_ax=survival_ax)
    fig.savefig(f"figures/fit-exponent-{static.value}-{temporal.value}-{deg_id}.pdf",
        bbox_inches='tight', pad_inches=0.1)

    csv.writer(sys.stdout).writerow([static.value, temporal.value, opt_dt, min_s1, min_s2, min_s3, min_s4, alpha, delta])

