import json
import argparse
import sys
import itertools
import string

from scipy import interpolate
from scipy import stats
import numpy as np
import tqdm
import matplotlib.pyplot as plt

import matplotlib 
matplotlib.rc('legend', fontsize='small', handlelength=1, markerscale=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

from network_configs import StaticModel, TemporalModel, ProbDist 

def all_dts(single_summaries, homogeneous_summaries, prob_dist):
  dts = []
  for name, summaries in [("single", single_summaries), ("homogeneous", homogeneous_summaries)]:
    sizes = sorted(summaries, key=int)
    for size_s in sizes:
      s = summaries[size_s][prob_dist.value]
      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts.append(set(keys))
      print(f"{name} {size_s} dtcs:", keys, file=sys.stderr)

  return set.intersection(*dts)

def fit_exponent(single_summaries, homogeneous_summaries, prob_dist, dtc, nu_para, nu_perp, dims, ax):
  max_size_s = max(single_summaries, key=int)
  size = int(max_size_s)
  s = single_summaries[max_size_s][prob_dist.value]
  time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])
  mk_survival = np.array(s[str(dtc)]["km-survival-dist"][0])
  max_t = 0.04*size**(nu_para[dims]/(nu_perp[dims]*dims))
  min_t = 2e0
  mask = (time > min_t) & (time < max_t)
  res = stats.linregress(np.log(time[mask]), np.log(mk_survival[mask]))
  delta = -res.slope
  ax.plot(time, mk_survival, ls='--', marker='o', label="$\\hat{P}(t)$", ms=2)
  ax.plot(time, np.exp(res.intercept)*time**(-delta), ls='-', marker='')

  max_size_s = max(homogeneous_summaries, key=int)
  size = int(max_size_s)
  s = homogeneous_summaries[max_size_s][prob_dist.value]
  time = np.array(s[str(dtc)]["time-bin-edges"][0][:-1])
  density = np.array(s[str(dtc)]["density"][0])
  max_t = 0.04*size**(nu_para[dims]/(nu_perp[dims]*dims))
  min_t = 2e0
  mask = (time > min_t) & (time < max_t)
  res = stats.linregress(np.log(time[mask]), np.log(density[mask]))
  alpha = -res.slope
  ax.plot(time, density, ls='--', marker='^', label="$\\rho(t)$", ms=2)
  ax.plot(time, np.exp(res.intercept)*time**(-alpha), ls='--', marker='')
    
  return alpha, delta

def fit_collapse(single_summaries, homogeneous_summaries, prob_dist, dtc, beta, beta_p, nu_para, nu_perp, dims, max_size=np.inf):
  gamma = nu_para + dims*nu_perp - beta - beta_p
  upsilon = dims*nu_perp - beta_p
  try:
    sizes = sorted(single_summaries, key=int)
    sizes = [s for s in sizes if int(s) <= max_size]

    interp = {}
    times = {}
    mk_survivals = {}

    time_min = -np.inf
    time_max = np.inf

    for size_s in sizes:
      s = single_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - float(dtc))

      mask = (np.array(s[dtc]["km-survival-dist"][0]) > 0)

      scaled_time = np.array(s[dtc]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims))
      scaled_mk_survival = np.array(s[dtc]["km-survival-dist"][0])*size**(beta_p/(nu_perp*dims))

      time_min = max(time_min, np.amin(scaled_time[mask]))
      time_max = min(time_max, np.amax(scaled_time[mask]))
      interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_mk_survival[mask]))
      times[size_s] = scaled_time[mask]
      mk_survivals[size_s] = scaled_mk_survival[mask]

    survival_msd = 0.0
    mean_survival = lambda t: np.mean([interp[s](t) for s in interp])
    ssd = 0.0
    count = 0
    for s in sizes:
      for t, p in zip(times[s], mk_survivals[s]):
        if t >= time_min and t <= time_max:
          ssd += (mean_survival(np.log(t)) - np.log(p))**2
          count += 1
    if count > 0:
      survival_msd += ssd/count
    else:
      survival_msd += np.inf

    sizes = sorted(homogeneous_summaries, key=int)
    sizes = [s for s in sizes if int(s) <= max_size]

    interp = {}
    times = {}
    densities = {}

    time_min = -np.inf
    time_max = np.inf

    for size_s in sizes:
      s = homogeneous_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - float(dtc))

      mask = (np.array(s[dtc]["density"][0]) > 0)

      scaled_time = np.array(s[dtc]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims))
      scaled_density = np.array(s[dtc]["density"][0])*size**(beta/(nu_perp*dims))

      if len(scaled_time[mask]) == 0:
        continue

      time_min = max(time_min, np.amin(scaled_time[mask]))
      time_max = min(time_max, np.amax(scaled_time[mask]))
      interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_density[mask]))
      times[size_s] = scaled_time[mask]
      densities[size_s] = scaled_density[mask]

    density_msd = 0.0
    mean_density = lambda t: np.mean([interp[s](t) for s in interp])
    count = 0
    ssd = 0.0
    for s in sizes:
      if s in times and s in densities:
        for t, d in zip(times[s], densities[s]):
          if t >= time_min and t <= time_max:
            ssd += (mean_density(np.log(t)) - np.log(d))**2
            count += 1
    if count > 0:
      density_msd = ssd/count
    else:
      density_msd = np.inf


    sizes = sorted(single_summaries, key=int)
    sizes = [s for s in sizes if int(s) <= max_size]

    interp = {}
    times = {}
    masses = {}

    time_min = -np.inf
    time_max = np.inf

    for size_s in sizes:
      s = single_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - float(dtc))

      mask = (np.array(s[dtc]["mass-times-to-end-dist"][0]) > 0)

      scaled_time = np.array(s[dtc]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims))
      scaled_masses = np.array(s[dtc]["mass-times-to-end-dist"][0])*size**(-gamma/(nu_perp*dims))

      time_min = max(time_min, np.amin(scaled_time[mask]))
      time_max = min(time_max, np.amax(scaled_time[mask]))
      interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_masses[mask]))
      times[size_s] = scaled_time[mask]
      masses[size_s] = scaled_masses[mask]

    mass_msd = 0.0
    mean_mass = lambda t: np.mean([interp[s](t) for s in interp])
    ssd = 0.0
    count = 0
    for s in sizes:
      for t, p in zip(times[s], masses[s]):
        if t >= time_min and t <= time_max:
          ssd += (mean_mass(np.log(t)) - np.log(p))**2
          count += 1
    if count > 0:
      mass_msd += ssd/count
    else:
      mass_msd += np.inf


    sizes = sorted(single_summaries, key=int)
    sizes = [s for s in sizes if int(s) <= max_size]

    interp = {}
    times = {}
    volumes = {}

    time_min = -np.inf
    time_max = np.inf

    for size_s in sizes:
      s = single_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - float(dtc))

      mask = (np.array(s[dtc]["volume-times-to-end-dist"][0]) > 0)

      scaled_time = np.array(s[dtc]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims))
      scaled_volumes = np.array(s[dtc]["volume-times-to-end-dist"][0])*size**(-upsilon/(nu_perp*dims))

      time_min = max(time_min, np.amin(scaled_time[mask]))
      time_max = min(time_max, np.amax(scaled_time[mask]))
      interp[size_s] = interpolate.interp1d(np.log(scaled_time[mask]), np.log(scaled_volumes[mask]))
      times[size_s] = scaled_time[mask]
      volumes[size_s] = scaled_volumes[mask]

    volumes_msd = 0.0
    mean_volume = lambda t: np.mean([interp[s](t) for s in interp])
    ssd = 0.0
    count = 0
    for s in sizes:
      for t, p in zip(times[s], volumes[s]):
        if t >= time_min and t <= time_max:
          ssd += (mean_volume(np.log(t)) - np.log(p))**2
          count += 1
    if count > 0:
      volumes_msd += ssd/count
    else:
      volumes_msd += np.inf

    return np.sqrt(survival_msd), np.sqrt(density_msd), np.sqrt(mass_msd), np.sqrt(volumes_msd)
  except ValueError as e:
    return np.inf, np.inf, np.inf, np.inf

if __name__=="__main__":
  comb = [
    (StaticModel.erdos, TemporalModel.poisson, 8, 4),
    (StaticModel.regular, TemporalModel.bursty_280, 9, 4),
    (StaticModel.erdos, TemporalModel.self_exciting_02_08_10, 8, 4),
    (StaticModel.grid_1, TemporalModel.poisson, 2, 1)
  ]

  beta =    {1: 0.276, 2: 0.583, 3: 0.813, 4: 1.0}
  nu_perp = {1: 1.096, 2: 0.733, 3: 0.584, 4: 0.5}
  nu_para = {1: 1.734, 2: 1.295, 3: 1.11,  4: 1.0}

  min_dt = 0
  max_dt = 2
  max_collapse_size = np.inf

  rms_fig, rms_axes = plt.subplots(ncols=2, nrows=2, figsize=(6.4, 6.4),
  			   gridspec_kw={'wspace': 0.3, 'hspace': 0.4})
  rms_axes = list(itertools.chain(*rms_axes))

  exponent_fig, exponent_axes = plt.subplots(ncols=2, nrows=2, figsize=(6.4, 6.4),
  			   gridspec_kw={'wspace': 0.3, 'hspace': 0.4})
  exponent_axes = list(itertools.chain(*exponent_axes))

  for (static, temporal, deg_id, dims), rms_ax, exp_ax, letter in zip(comb, rms_axes, exponent_axes, string.ascii_lowercase):
    prob_dist = ProbDist.deterministic
    if static is StaticModel.grid_1:
      max_collapse_size = 2**12
    else:
      max_collapse_size = np.inf

    print("plotting:", static, temporal, dims, file=sys.stderr)

    single_filename = f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-single.json"
    homogeneous_filename = f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-homogeneous.json"

    with open(single_filename) as sngl_f, open(homogeneous_filename) as homo_f:
      single_summaries = json.load(sngl_f)
      homogeneous_summaries = json.load(homo_f)

      dtc_candidates = all_dts(single_summaries, homogeneous_summaries, prob_dist)
      dtc_candidates = [c for c in sorted(dtc_candidates) if float(c) >= min_dt and float(c) <= max_dt]

      print("dtcs:", dtc_candidates, file=sys.stderr)

      survival_ssd = []
      density_ssd = []
      mass_ssd = []
      volume_ssd = []
      for k in tqdm.tqdm(dtc_candidates):
        s1, s2, s3, s4 = fit_collapse(single_summaries, homogeneous_summaries, prob_dist, dtc=k,
            beta=beta[dims], beta_p=beta[dims], nu_perp=nu_perp[dims], nu_para=nu_para[dims], dims=dims, max_size=max_collapse_size)
        survival_ssd.append(s1)
        density_ssd.append(s2)
        mass_ssd.append(s3)
        volume_ssd.append(s4)
  
      opt_dt, min_s1, min_s2, min_s3, min_s4 = min(zip(dtc_candidates, survival_ssd, density_ssd, mass_ssd, volume_ssd), key=lambda a: a[1] + a[2] + a[3] + a[4])
  
      rms_ax.axvline(float(opt_dt), ls='--', color='gray')
      rms_ax.text(x=0.5, y=0.95, s=f"$\\delta t = {opt_dt}$", fontsize='smaller',
        verticalalignment='top', horizontalalignment='left', rotation=270,
        transform=rms_ax.transAxes)
      rms_ax.plot(list(map(float, dtc_candidates)), survival_ssd, label="$\\hat{P}(t)$", marker='.')
      rms_ax.plot(list(map(float, dtc_candidates)), density_ssd, label="$\\rho(t)$", marker='v')
      rms_ax.plot(list(map(float, dtc_candidates)), mass_ssd, label="$M(t)$", marker='h')
      rms_ax.plot(list(map(float, dtc_candidates)), volume_ssd, label="$V(t)$", marker='^')

      delta = 6e-3
      if static is StaticModel.grid_1:
        delta = 1e-1

      rms_ax.set_xlim(float(opt_dt)-delta, float(opt_dt)+delta)
      rms_ax.set_xlabel("$\\delta t$")
      rms_ax.set_ylabel("RMS Error w/ mean curve")
      rms_ax.set_yscale("log")
      rms_ax.legend(loc='lower left')
      rms_ax.text(-0.2, 0.95, "("+letter+")", transform=rms_ax.transAxes)
      rms_ax.tick_params(axis='x', labelsize='smaller')
      rms_ax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%g'))

      alpha, delta = fit_exponent(single_summaries, homogeneous_summaries, prob_dist,
          dtc=float(opt_dt), nu_para=nu_para, nu_perp=nu_perp, dims=dims,
          ax=exp_ax)

      exp_ax.set_xscale("log")
      exp_ax.set_yscale("log")
      exp_ax.set_xlabel("Time")
      exp_ax.legend()
      exp_ax.text(-0.25, 0.95, "("+letter+")", transform=exp_ax.transAxes)
      exp_ax.set_ylim(1e-3, 2e0)
      if static is StaticModel.grid_1:
        exp_ax.set_ylim(1e-1, 2e0)


  rms_fig.savefig(f"figures/fig-example-rmse.pdf",
        bbox_inches='tight', pad_inches=0.1)
  exponent_fig.savefig(f"figures/fig-example-exponent.pdf",
        bbox_inches='tight', pad_inches=0.1)
