import json
import string
import enum
from collections import defaultdict

import numpy as np
import matplotlib.pyplot as plt

import matplotlib 
matplotlib.rc('legend', fontsize=8, handlelength=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)

class RealNetworks(enum.Enum):
  air = "air"
  transport = "transport"
  twitter = "twitter"
  mobile = "mobile"


fig, axes = plt.subplots(4, 4, squeeze=False, sharex='col',
                           figsize=(3.375*2*1.5, 2.08*2*1.5),
                           gridspec_kw={'wspace': 0.5, 'hspace': 0.3})

nets = [RealNetworks.air, RealNetworks.transport, RealNetworks.twitter, RealNetworks.mobile]

for i, real in enumerate(nets):
  dtcs = 0
  size = 0
  if real is RealNetworks.air:
    dt_unit_name = "minutes"
    dt_unit = 60
    size = 279
    dtcs = [470]
  elif real is RealNetworks.transport:
    dt_unit_name = "minutes"
    dt_unit = 60
    size = 6986
    dtcs = [670]
  elif real is RealNetworks.twitter:
    dt_unit_name = "hours"
    dt_unit = 3600
    size = 17313623
    dtcs = [5500, 79200]
  elif real is RealNetworks.mobile:
    dt_unit_name = "hours"
    dt_unit = 3600
    size = 5193086
    dtcs = [27000, 50400]

  dtc_colors = ["tab:grey", "tab:olive", "tab:cyan"]

  with open(f"networks/{real.name}/single.json") as f:
    singles = json.load(f)[str(size)]["deterministic"]
    keys = sorted(singles.keys(), key=lambda k: singles[k]["dt"][0])
    dts = np.array([singles[k]["dt"][0] for k in keys])/dt_unit

    mass_means = np.array([singles[k]["mean-mass"][0] for k in keys])
    mass_medians = np.array([singles[k]["mean-mass"][2] for k in keys])
    mass_sems = np.array([singles[k]["mean-mass"][1] for k in keys])
    axes[0][i].errorbar(dts, mass_means, yerr=mass_sems*1.96)
    # axes[0][i].plot(dts, mass_medians)
    for dtc, col in zip(dtcs, dtc_colors):
      axes[0][i].axvline(x=dtc/dt_unit, ls='--', color=col)
    axes[0][i].ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

    axes[0][i].text(-0.2, 1.2, "("+string.ascii_lowercase[i]+")", transform=axes[0][i].transAxes)


    volume_means = np.array([singles[k]["mean-volume"][0] for k in keys])
    volume_medians = np.array([singles[k]["mean-volume"][2] for k in keys])
    volume_sems = np.array([singles[k]["mean-volume"][1] for k in keys])
    axes[1][i].errorbar(dts, volume_means, yerr=volume_sems*1.96)
    # axes[1][i].plot(dts, volume_medians)
    for dtc, col in zip(dtcs, dtc_colors):
      axes[1][i].axvline(x=dtc/dt_unit, ls='--', color=col)
    axes[1][i].ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

  with open(f"networks/{real.name}/homogeneous-alt.json") as f:
    homogeneous = json.load(f)[str(size)]["deterministic"]
    keys = sorted(homogeneous.keys(), key=lambda k: homogeneous[k]["dt"][0])
    dts = np.array([homogeneous[k]["dt"][0] for k in keys])/dt_unit

    static_density_means = np.array([homogeneous[k]["dt-band-static-density"][0] for k in keys])
    static_density_medians = np.array([homogeneous[k]["dt-band-static-density"][2] for k in keys])
    static_density_sems = np.array([homogeneous[k]["dt-band-static-density"][1] for k in keys])
    axes[2][i].errorbar(dts, static_density_means, yerr=static_density_sems*1.96)
    # axes[2][i].plot(dts, static_density_medians)
    for dtc, col in zip(dtcs, dtc_colors):
      axes[2][i].axvline(x=dtc/dt_unit, ls='--', color=col)
    axes[2][i].ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

    susceptibility_means = np.array([homogeneous[k]["dt-band-susceptibility"][0] for k in keys])
    susceptibility_medians = np.array([homogeneous[k]["dt-band-susceptibility"][2] for k in keys])
    susceptibility_sems = np.array([homogeneous[k]["dt-band-susceptibility"][1] for k in keys])
    axes[3][i].errorbar(dts, susceptibility_means, yerr=susceptibility_sems)
    # axes[3][i].plot(dts, susceptibility_medians)
    for dtc, col in zip(dtcs, dtc_colors):
      axes[3][i].axvline(x=dtc/dt_unit, ls='--', color=col)
    axes[3][i].ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

  axes[3][i].set_xlabel(f"$\\delta t$ ({dt_unit_name})")


axes[0][0].set_ylabel("Mass $M$")
axes[1][0].set_ylabel("Volume $V$")
axes[2][0].set_ylabel("$\\rho_{stat}$")
axes[3][0].set_ylabel("$\\chi$")

fig.savefig("figures/real.pdf", bbox='tight', pad_inches=0)
