import json
import string
import enum
import os
from collections import defaultdict

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.ticker import FuncFormatter

import matplotlib 
matplotlib.rc('legend', fontsize='xx-small', handlelength=1, markerscale=1)
matplotlib.rc('lines', linewidth=0.75, markersize=2)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

class RealNetworks(enum.Enum):
  air = "air"
  transport = "transport"
  twitter = "twitter"
  mobile = "mobile"

def to_tex_sci(num):
  s, e =  np.format_float_scientific(num, precision=1, exp_digits=1).split('e')
  if int(e) == 0:
    return float(s)
  else:
    return f"{float(s)} \\times 10^{{{int(e)}}}"


fig, axes = plt.subplots(4, 2, figsize=(3.375*1.5, 4.5*1.5),
                           gridspec_kw={'wspace': 0.5, 'hspace': 0.6, 'top': 0.95, 'bottom': 0.05, 'right': 0.95})

nets = [RealNetworks.air, RealNetworks.transport, RealNetworks.twitter, RealNetworks.mobile]

axes = axes.flatten()

for i, real in enumerate(nets):
  dtcs = 0
  size = 0
  if real is RealNetworks.air:
    dt_unit_name = "minutes"
    dt_unit = 60
    size = 279
    dtcs = [470]
    dt_lims = (1.0, 40)
  elif real is RealNetworks.transport:
    dt_unit_name = "minutes"
    dt_unit = 60
    size = 6986
    dtcs = [670]
    dt_lims = (1.0, 40)
  elif real is RealNetworks.twitter:
    dt_unit_name = "hours"
    dt_unit = 3600
    size = 17313623
    dtcs = [1500, 5500, 79200]
    dt_lims = (.0001, 70)
  elif real is RealNetworks.mobile:
    dt_unit_name = "hours"
    dt_unit = 3600
    size = 5193086
    dtcs = [27000, 50400]
    dt_lims = (.01, 96)

  dtc_colors = ["tab:grey", "tab:olive", "tab:cyan"]

  with open(f"networks/{real.name}/single.json") as f:
    singles = json.load(f)[str(size)]["deterministic"]
    keys = sorted(singles.keys(), key=lambda k: singles[k]["dt"][0])
    dts = np.array([singles[k]["dt"][0] for k in keys])/dt_unit

    mass_means = np.array([singles[k]["mean-mass"][0] for k in keys])
    mass_sems = np.array([singles[k]["mean-mass"][1] for k in keys])
    mass_max = np.max(mass_means)
    axes[i].errorbar(dts,
        mass_means/mass_max,
        yerr=mass_sems*1.96/mass_max,
        label=f"$M/{to_tex_sci(mass_max)}$")

    volume_means = np.array([singles[k]["mean-volume"][0] for k in keys])
    volume_sems = np.array([singles[k]["mean-volume"][1] for k in keys])
    volume_max = np.max(volume_means)
    axes[i].errorbar(dts,
        volume_means/volume_max,
        yerr=volume_sems*1.96/volume_max,
        label=f"$V/{to_tex_sci(volume_max)}$")

  with open(f"networks/{real.name}/homogeneous-alt.json") as f:
    homogeneous = json.load(f)[str(size)]["deterministic"]
    keys = sorted(homogeneous.keys(), key=lambda k: homogeneous[k]["dt"][0])
    dts = np.array([homogeneous[k]["dt"][0] for k in keys])/dt_unit

    static_density_means = np.array([homogeneous[k]["dt-band-static-density"][0] for k in keys])
    static_density_sems = np.array([homogeneous[k]["dt-band-static-density"][1] for k in keys])
    static_density_max = np.max(static_density_means)
    axes[i].errorbar(dts,
        static_density_means/static_density_max,
        yerr=static_density_sems*1.96/static_density_max,
        label=f"$\\rho_{{stat}}/{to_tex_sci(static_density_max)}$")

    susceptibility_means = np.array([homogeneous[k]["dt-band-susceptibility"][0] for k in keys])
    susceptibility_sems = np.array([homogeneous[k]["dt-band-susceptibility"][1] for k in keys])
    susceptibility_max = np.max(susceptibility_means)
    axes[i].errorbar(dts,
        susceptibility_means/susceptibility_max,
        yerr=susceptibility_sems*1.96/susceptibility_max,
        label=f"$\\chi/{to_tex_sci(susceptibility_max)}$")

  for dtc, col in zip(dtcs, dtc_colors):
    axes[i].axvline(x=dtc/dt_unit, ls='--', color=col)

  axes[i].legend()
  axes[i].set_xlim(dt_lims)
  axes[i].set_xlabel(f"$\\delta t$ ({dt_unit_name})")
  axes[i].text(-0.1, 1.05, "("+string.ascii_lowercase[i]+")", transform=axes[i].transAxes)
  axes[i].set_xscale("log")

for network, dt, ax, letter in zip(
    ["air", "transport", "twitter", "mobile"],
    [470, 670, 5500, 50400], axes[4:], string.ascii_lowercase[4:]):
  filename = os.path.join("networks", network, "out-comp-sizes.jsonl")
  with open(filename) as f:
    for l in f:
      j = json.loads(l)
      if np.isclose(j['dt'], dt):
        if network == "air":
          time_unit = "(days)"
          time_unit_secs = 3600*24
          ticks = np.array([1, 3, 5, 7, 9])*time_unit_secs
        elif network == "transport":
          time_unit = "(hours)"
          time_unit_secs = 3600
          ticks = np.array([2, 8, 14, 20])*time_unit_secs
        elif network == "twitter":
          time_unit = "(days)"
          time_unit_secs = 3600*24
          ticks = np.array([10, 70, 130, 190, 250])*time_unit_secs
        elif network == "mobile":
          time_unit = "(days)"
          time_unit_secs = 3600*24
          ticks = np.array([20, 40, 60, 80, 100])*time_unit_secs
        ax.set_xlabel("Event time " + time_unit)

        volume_aspect = abs((j["time-bin-edges"][-1] - j["time-bin-edges"][0])/\
                             (j["volume-bin-edges"][-1] - j["volume-bin-edges"][0]))/\
                             ((len(j["time-bin-edges"])-1)/(len(j["volume-bin-edges"])-1))
        ax.imshow(np.transpose(j["volume-bin-counts"]) > 0, aspect=volume_aspect,
                         origin="lower", cmap="binary", vmin=0, vmax=1, interpolation='none',
                         extent=[0, (j["time-bin-edges"][-1] - j["time-bin-edges"][0]),
                                   j["volume-bin-edges"][0], j["volume-bin-edges"][-1]])
        yticks = np.array([0., 0.5, 1.0])
        ax.set_yticks(yticks*j["volume-bin-edges"][-1])
        ax.set_yticklabels([f"{t}" for t in yticks])
        ax.set_ylabel("Fractional Volume")
        # ax.ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        formatter = lambda t: f"{t/time_unit_secs:g}"
        ax.set_xticks(ticks)
        ax.set_xticklabels(map(formatter, ticks))

        ax.text(-0.1, 1.25, f"({letter})", transform=ax.transAxes)

fig.savefig("figures/real-condensed.pdf", bbox='tight', pad_inches=0)
