import json
import string
import enum
import argparse
from collections import defaultdict

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from network_configs import StaticModel, TemporalModel, ProbDist

import matplotlib 
matplotlib.rc('legend', fontsize='x-small', handlelength=1, markerscale=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

def plot_system(static, temporal, prob_dist, deg_id, dims, dtc_s,
                  beta, beta_p, nu_para, nu_perp,
                  mass_ax, volume_ax, static_density_ax, susceptibility_ax,
                  mk_survival_ax, density_time_ax, mass_time_ax, volume_time_ax):

  dtc = float(dtc_s)

  delta = beta_p/nu_para
  theta = dims*nu_para - beta - beta_p
  gamma = nu_para + dims*nu_perp - beta - beta_p
  upsilon = dims*nu_perp - beta_p

  markers = [".", "v", "h", "^", "*", ">", "2", "<", "d", "+", "x"]
  
  with open(f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-single.json") as f:
    single_summaries = json.load(f)
    sizes = sorted(single_summaries, key=int)

    for size_s, marker in zip(sizes, markers):
      s = single_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - dtc)

      mk_survival_ax.errorbar(
          np.array(s[dtc_s]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims)),
          np.array(s[dtc_s]["km-survival-dist"][0])*size**(beta_p/(nu_perp*dims)),
          yerr=np.array(s[dtc_s]["km-survival-dist"][1])*1.96*size**(beta_p/(nu_perp*dims)),
          label=f"$N=2^{{{int(np.log2(size))}}}$", marker=marker, ls='-')

      mass_means = np.array([s[k]["mean-mass"][0] for k in keys])
      mass_ax.plot(
          -taus[dts - dtc < 0]*size**(1/(dims*nu_perp)),
          mass_means[dts - dtc < 0]*size**(-gamma/(dims*nu_perp)),
          label=f"N={size}", marker=marker, ls='-')

      mass_time_ax.errorbar(
          np.array(s[dtc_s]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims)),
          np.array(s[dtc_s]["mass-times-to-end-dist"][0])*size**(-gamma/(nu_perp*dims)),
          yerr=np.array(s[dtc_s]["mass-times-to-end-dist"][1])*1.96*size**(-gamma/(nu_perp*dims)),
          label=f"N={size}", marker=marker, ls='-')

      volume_means = np.array([s[k]["mean-volume"][0] for k in keys])
      volume_ax.plot(
          -taus[dts - dtc < 0]*size**(1/(dims*nu_perp)),
          volume_means[dts - dtc < 0]*size**(-upsilon/(dims*nu_perp)),
          label=f"N={size}", marker=marker, ls='-')

      volume_time_ax.errorbar(
          np.array(s[dtc_s]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims)),
          np.array(s[dtc_s]["volume-times-to-end-dist"][0])*size**(-upsilon/(nu_perp*dims)),
          yerr=np.array(s[dtc_s]["volume-times-to-end-dist"][1])*1.96*size**(-upsilon/(nu_perp*dims)),
          label=f"N={size}", marker=marker, ls='-')

  mk_survival_ax.set_xscale("log")
  mk_survival_ax.set_yscale("log")
  mk_survival_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$")
  mk_survival_ax.set_ylabel("$\\hat{P}(t) N^{\\beta'/(d \\nu_\\perp)}$")
  mk_survival_ax.legend(ncol=2, loc='lower left')

  mass_ax.set_xscale("log")
  mass_ax.set_yscale("log")
  mass_ax.set_xlabel("$(\\delta t_c - \\delta t) N^{1/(d \\nu_\\perp)}$")
  mass_ax.set_ylabel("$M N^{-\\gamma/(d \\nu_\\perp)}$")

  volume_ax.set_xscale("log")
  volume_ax.set_yscale("log")
  volume_ax.set_xlabel("$(\\delta t_c - \\delta t) N^{1/(d \\nu_\\perp)}$")
  volume_ax.set_ylabel("$V N^{-\\upsilon/(d \\nu_\\perp)}$")

  mass_time_ax.set_xscale("log")
  mass_time_ax.set_yscale("log")
  mass_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$")
  mass_time_ax.set_ylabel("$M(t) N^{-\\gamma/(d \\nu_\\perp)}$")

  volume_time_ax.set_xscale("log")
  volume_time_ax.set_yscale("log")
  volume_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$")
  volume_time_ax.set_ylabel("$V(t) N^{-\\upsilon/(d \\nu_\\perp)}$")

  with open(f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-homogeneous.json") as f:
    homogeneous_summaries = json.load(f)
    sizes = sorted(homogeneous_summaries, key=int)

    for size_s, marker in zip(sizes, markers):
      s = homogeneous_summaries[size_s][prob_dist.value]
      size = int(size_s)

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      dts = np.array([s[k]["dt"][0] for k in keys])
      taus = (dts - dtc)

      density_time_ax.errorbar(
          np.array(s[dtc_s]["time-bin-edges"][0][:-1])*size**(-nu_para/(nu_perp*dims)),
          np.array(s[dtc_s]["density"][0])*size**(beta/(nu_perp*dims)),
          yerr=np.array(s[dtc_s]["density"][1])*1.96*size**(beta/(nu_perp*dims)),
          label=f"N={size}", marker=marker, ls='-')

      static_density_means = np.array([s[k]["dt-band-static-density"][0] for k in keys])
      static_density_ax.plot(
          taus[dts - dtc > 0]*size**(1/(dims*nu_perp)),
          static_density_means[dts - dtc > 0]*size**(beta/(dims*nu_perp)),
          label=f"N={size}", marker=marker, ls='-')

      susceptibility_means = np.array([s[k]["dt-band-susceptibility"][0] for k in keys])
      susceptibility_ax.plot(
          taus[dts - dtc > 0]*size**(1/(dims*nu_perp)),
          susceptibility_means[dts - dtc > 0]*size**(-gamma/(dims*nu_perp)),
          label=f"N={size}", marker=marker, ls='-')

  density_time_ax.set_xscale("log")
  density_time_ax.set_yscale("log")
  density_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$")
  density_time_ax.set_ylabel("$\\rho(t) N^{\\beta'/(d \\nu_\\perp)}$")

  static_density_ax.set_xscale("log")
  static_density_ax.set_yscale("log")
  static_density_ax.set_xlabel("$(\\delta t - \\delta t_c) N^{1/(d \\nu_\\perp)}$")
  static_density_ax.set_ylabel("$\\rho_{stat} N^{\\beta/(d \\nu_\\perp)}$")

  susceptibility_ax.set_xscale("log")
  susceptibility_ax.set_yscale("log")
  susceptibility_ax.set_xlabel("$(\\delta t - \\delta t_c) N^{1/(d \\nu_\\perp)}$")
  susceptibility_ax.set_ylabel("$\\chi N^{-\\gamma/(d \\nu_\\perp)}$")

if __name__=="__main__":
  parser = argparse.ArgumentParser(description="Plots quantities of random networks")
  parser.add_argument('static', type=StaticModel)
  parser.add_argument('temporal', type=TemporalModel)
  parser.add_argument('dtc', type=str)
  args = parser.parse_args()

  beta =    {1: 0.276, 2: 0.583, 3: 0.814, 4: 1.0}
  nu_para = {1: 1.734, 2: 1.295, 3: 1.110, 4: 1.0}
  nu_perp = {1: 1.097, 2: 0.733, 3: 0.584, 4: 0.5}
  beta_p = beta

  static = args.static
  temporal = args.temporal
  prob_dist = ProbDist.deterministic

  print("plotting", static, temporal, prob_dist)

  dtc_s = args.dtc

  deg_id = 9
  dims = 4

  survival_time_lims = (1e-3, 1e0)

  mass_tau_lim = (1e-4, 3e1)
  mass_lim = (1e-4, 5e0)

  volume_tau_lim = (1e-4, 3e1)
  volume_lim = (1e-4, 5e0)

  density_time_lims = (1e-3, 1e0)
  static_density_tau_lims = (5e-2, 5e1)
  susceptibility_tau_lims = (1e-4, 5e1)

  if static is StaticModel.erdos:
    deg_id = 8
  elif static is StaticModel.grid_1 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-7, 5e0)

    mass_tau_lim = (5e-2, 5e3)
    mass_lim = (1e-9, 1e0)

    volume_tau_lim = (5e-2, 5e3)
    volume_lim = (1e-4, 5e0)

    density_time_lims = (1e-7, 5e0)
    static_density_tau_lims = (1e-2, 5e3)
    susceptibility_tau_lims = (1e-2, 5e3)

    deg_id = 2
    dims = 1
  elif static is StaticModel.grid_2 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 4
    dims = 2
  elif static is StaticModel.grid_3 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 6
    dims = 3
  elif static is StaticModel.grid_4 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 8
    dims = 4

  fig, axes = plt.subplots(2, 3,
  			   figsize=(3.375*2*1.5, 1.44*2*1.5),
  			   gridspec_kw={'wspace': 0.3, 'hspace': 0.4})

  ((mk_survival_ax, mass_ax, volume_ax),
    (density_time_ax, static_density_ax, susceptibility_ax)) = axes

  mass_time_ax = inset_axes(mass_ax, width="55%", height="45%", loc='lower left', bbox_to_anchor=(.2, .2, 1., 1.), bbox_transform=mass_ax.transAxes, borderpad=0)
  volume_time_ax = inset_axes(volume_ax, width="55%", height="45%", loc='lower left', bbox_to_anchor=(.2, .2, 1., 1.), bbox_transform=volume_ax.transAxes, borderpad=0)


  plot_system(static, temporal, prob_dist, deg_id, dims, dtc_s,
                beta[dims], beta_p[dims], nu_para[dims], nu_perp[dims],
                mass_ax, volume_ax, static_density_ax, susceptibility_ax,
                mk_survival_ax, density_time_ax, mass_time_ax, volume_time_ax)

  for ax, letter in zip(axes.flatten(), string.ascii_lowercase):
    ax.text(-0.25, 0.90, f"({letter})", transform=ax.transAxes)

  # mass_time_ax.xaxis.set_label_position("top")
  mass_time_ax.spines['top'].set_visible(True)
  mass_time_ax.set_ylabel("$M(t) N^{-\\gamma/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  mass_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  # mass_time_ax.yaxis.set_label_position("right")
  mass_time_ax.spines['right'].set_visible(True)
  mass_time_ax.tick_params(axis='both', which='both', direction='in', labelsize='x-small', pad=1)


  # volume_time_ax.xaxis.set_label_position("top")
  volume_time_ax.spines['top'].set_visible(True)
  volume_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  volume_time_ax.set_ylabel("$V(t) N^{-\\upsilon/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  # volume_time_ax.yaxis.set_label_position("right")
  volume_time_ax.spines['right'].set_visible(True)
  volume_time_ax.tick_params(axis='both', which='both', direction='in', labelsize='x-small', pad=1)

  mk_survival_ax.set_xlim(survival_time_lims)
  mass_ax.set_xlim(mass_tau_lim)
  mass_ax.set_ylim(mass_lim)
  volume_ax.set_xlim(volume_tau_lim)
  volume_ax.set_ylim(volume_lim)

  density_time_ax.set_xlim(density_time_lims)
  susceptibility_ax.set_xlim(susceptibility_tau_lims)
  static_density_ax.set_xlim(static_density_tau_lims)

  fig.savefig(f"figures/random-{static.value}-{temporal.value}-{deg_id}.pdf",
      bbox_inches='tight', pad_inches=0.1)

