import json
import string
import enum
import argparse
import os
from collections import defaultdict

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.ticker import FuncFormatter

from network_configs import StaticModel, TemporalModel, ProbDist

from plot_scaling import plot_system

import matplotlib 
matplotlib.rc('legend', fontsize='x-small', handlelength=1, markerscale=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

if __name__=="__main__":
  parser = argparse.ArgumentParser(description="Plots quantities of random networks")
  parser.add_argument('static', type=StaticModel)
  parser.add_argument('temporal', type=TemporalModel)
  parser.add_argument('dtc', type=str)
  args = parser.parse_args()

  beta =    {1: 0.276, 2: 0.583, 3: 0.813, 4: 1.0}
  nu_perp = {1: 1.096, 2: 0.733, 3: 0.584, 4: 0.5}
  nu_para = {1: 1.734, 2: 1.295, 3: 1.11, 4: 1.0}
  beta_p = beta

  static = args.static
  temporal = args.temporal
  prob_dist = ProbDist.deterministic

  print("plotting", static, temporal, prob_dist)

  dtc_s = args.dtc

  deg_id = 9
  dims = 4

  survival_time_lims = (1e-3, 1e0)

  mass_tau_lim = (1e-4, 3e1)
  mass_lim = (1e-4, 5e0)

  volume_tau_lim = (1e-4, 3e1)
  volume_lim = (1e-4, 5e0)

  density_time_lims = (1e-3, 1e0)
  static_density_tau_lims = (5e-2, 5e1)
  susceptibility_tau_lims = (1e-4, 5e1)

  if static is StaticModel.erdos:
    deg_id = 8
  elif static is StaticModel.grid_1 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-7, 5e0)

    mass_tau_lim = (5e-2, 5e3)
    mass_lim = (1e-9, 1e0)

    volume_tau_lim = (5e-2, 5e3)
    volume_lim = (1e-4, 5e0)

    density_time_lims = (1e-7, 5e0)
    static_density_tau_lims = (1e-2, 5e3)
    susceptibility_tau_lims = (1e-2, 5e3)

    deg_id = 2
    dims = 1
  elif static is StaticModel.grid_2 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 4
    dims = 2
  elif static is StaticModel.grid_3 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 6
    dims = 3
  elif static is StaticModel.grid_4 and temporal is TemporalModel.poisson:
    survival_time_lims = (1e-5, 5e0)

    mass_tau_lim = (1e-2, 1e3)
    mass_lim = (1e-7, 5e0)

    volume_tau_lim = (1e-2, 1e3)
    volume_lim = (1e-4, 3e0)

    density_time_lims = (1e-5, 5e0)
    static_density_tau_lims = (1e-2, 1e2)
    susceptibility_tau_lims = (1e-2, 1e2)

    deg_id = 8
    dims = 4

  fig, axes = plt.subplots(4, 2,
  			   figsize=(3.375*1.5, 5.50*1.5),
  			   gridspec_kw={'wspace': 0.4, 'hspace': 0.35, 'top':0.95, 'bottom': 0.05, 'right': 0.95})

  mass_ax, volume_ax, mk_survival_ax, \
    density_time_ax, static_density_ax, susceptibility_ax, out_comp_before_ax, out_comp_after_ax = axes.flatten()

  mass_time_ax = inset_axes(mass_ax, width="50%", height="40%", loc='lower left', bbox_to_anchor=(.25, .2, 1., 1.), bbox_transform=mass_ax.transAxes, borderpad=0)
  volume_time_ax = inset_axes(volume_ax, width="50%", height="40%", loc='lower left', bbox_to_anchor=(.25, .2, 1., 1.), bbox_transform=volume_ax.transAxes, borderpad=0)


  plot_system(static, temporal, prob_dist, deg_id, dims, dtc_s,
                beta[dims], beta_p[dims], nu_para[dims], nu_perp[dims],
                mass_ax, volume_ax, static_density_ax, susceptibility_ax,
                mk_survival_ax, density_time_ax, mass_time_ax, volume_time_ax)

  for ax, letter in zip(axes.flatten()[:-2], string.ascii_lowercase):
    ax.text(-0.25, 0.95, f"({letter})", transform=ax.transAxes)

  for ax, letter in zip(axes.flatten()[-2:], string.ascii_lowercase):
    ax.text(-0.25, 1.2, f"({letter})", transform=ax.transAxes)

  # mass_time_ax.xaxis.set_label_position("top")
  mass_time_ax.spines['top'].set_visible(True)
  mass_time_ax.set_ylabel("$M(t) N^{-\\gamma/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  mass_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  # mass_time_ax.yaxis.set_label_position("right")
  mass_time_ax.spines['right'].set_visible(True)
  mass_time_ax.tick_params(axis='both', which='both', direction='in', labelsize='x-small', pad=1)


  # volume_time_ax.xaxis.set_label_position("top")
  volume_time_ax.spines['top'].set_visible(True)
  volume_time_ax.set_xlabel("$t N^{-\\nu_\\parallel/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  volume_time_ax.set_ylabel("$V(t) N^{-\\upsilon/(d \\nu_\\perp)}$", fontsize='small', labelpad=1)
  # volume_time_ax.yaxis.set_label_position("right")
  volume_time_ax.spines['right'].set_visible(True)
  volume_time_ax.tick_params(axis='both', which='both', direction='in', labelsize='x-small', pad=1)

  mk_survival_ax.set_xlim(survival_time_lims)
  mass_ax.set_xlim(mass_tau_lim)
  mass_ax.set_ylim(mass_lim)
  volume_ax.set_xlim(volume_tau_lim)
  volume_ax.set_ylim(volume_lim)

  density_time_ax.set_xlim(density_time_lims)
  susceptibility_ax.set_xlim(susceptibility_tau_lims)
  static_density_ax.set_xlim(static_density_tau_lims)

  # plot final scatter panels
  # TODO: proper filename format incorperating temporal and deg_id
  out_comp_filename = f"networks/random_large_networks/out-comp/{static.value}.jsonl"
  out_comp_dts = [0.07, 0.092]
  if os.path.exists(out_comp_filename):
    with open(out_comp_filename) as f:
      for ax, dt in zip([out_comp_before_ax, out_comp_after_ax], out_comp_dts):
        f.seek(0)
        for l in f:
          j = json.loads(l)
          if np.isclose(j['dt'], dt):
            volume_aspect = abs((mdates.epoch2num(j["time-bin-edges"][-1]) - mdates.epoch2num(j["time-bin-edges"][0]))/\
                                 (j["volume-bin-edges"][-1] - j["volume-bin-edges"][0]))/((len(j["time-bin-edges"])-1)/\
                                 (len(j["volume-bin-edges"])-1))
            ax.imshow(np.transpose(j["volume-bin-counts"]) > 0, aspect=volume_aspect,
                             origin="lower", cmap="binary", vmin=0, vmax=1, interpolation='none',
                             extent=[mdates.epoch2num(j["time-bin-edges"][0]), mdates.epoch2num(j["time-bin-edges"][-1]),
                                       j["volume-bin-edges"][0], j["volume-bin-edges"][-1]])
        ax.set_ylabel("Fractional Volume", labelpad=9.8)
        yticks = np.array([0., 0.5, 1.0])
        ax.set_yticks(yticks*j["volume-bin-edges"][-1])
        ax.set_yticklabels([f"{t}" for t in yticks])
        # ax.ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0,0))

        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        locator = mdates.AutoDateLocator(minticks=3, maxticks=5)

        formatter = FuncFormatter(lambda t, _: int((t - mdates.epoch2num(j["time-bin-edges"][0]))*24*3600 + 0.5))
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)
        ax.set_xlabel("Event time")

  fig.savefig(f"figures/random-{static.value}-{temporal.value}-{deg_id}-vertical.pdf",
      bbox_inches='tight', pad_inches=0.1)

