import enum
import glob
from itertools import product, chain
import os.path
import json
import fileinput
import re
from collections import defaultdict

import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from tqdm import tqdm

import matplotlib 
matplotlib.rc('legend', fontsize=8, handlelength=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)

class StaticModel(enum.Enum):
  erdos = "erdos"
  barabasi = "barabasi"
  heavytail_low = "heavytail_low"
  heavytail_medium = "heavytail_medium"
  heavytail_high = "heavytail_high"
  heavytail_veryhigh = "heavytail_veryhigh"
  heavytail_veryveryhigh = "heavytail_veryveryhigh"
  regular = "regular"
  grid = "grid"
class TemporalModel(enum.Enum):
  poisson = "poisson"
  bursty_low = "bursty_low"
  bursty_medium = "bursty_medium"
  bursty_high = "bursty_high"
  bursty_veryhigh = "bursty_veryhigh"
  bursty_veryveryhigh = "bursty_veryveryhigh"
class ProbDist(enum.Enum):
  deterministic = "deterministic"
  exponential = "exponential"


def plot_collapse_vs_time(
    summaries, static_model, temporal_model, prob_dist_type, deg_id,
    dtcs, beta, v_para, v_perp, d):
  z = v_para/v_perp
  beta_p = beta
  alpha = beta/v_para
  delta = beta_p/v_para
  theta = (d*v_perp - beta - beta_p)/v_para
  gamma = v_para + d*v_perp - beta - beta_p
  upsilon = d*v_perp - beta_p

  bins = None  # will be set from the first summary

  sizes = sorted(summaries, key=int)

  mk_survival_fig, mk_survival_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  mk_survival_scaling_fig, mk_survival_scaling_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))

  mass_fig, mass_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  mass_scaling_fig, mass_scaling_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))

  volume_fig, volume_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  volume_scaling_fig, volume_scaling_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))


  for size, marker in zip(sizes, [".", "v", "h", "^", "*", ">", "2", "<", "d"]):
    s = summaries[size][prob_dist_type.name]
    dtc = str(dtcs[int(size)])

    t_max = s[dtc]["window"][0][1]

    mk_survival_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1]),
        np.array(s[dtc]["km-survival-dist"][0]),
        yerr=np.array(s[dtc]["km-survival-dist"][1])*1.96,
        marker=marker, ls='-',
        label=f"N={size}")
    mk_survival_scaling_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1])*int(size)**(-v_para/(v_perp*d)),
        np.array(s[dtc]["km-survival-dist"][0])*int(size)**(delta/(v_perp*d)),
        yerr=np.array(s[dtc]["km-survival-dist"][1])*1.96*int(size)**(delta/(v_perp*d)),
        marker=marker, ls='-',
        label=f"N={size}")

    volume_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1]),
        np.array(s[dtc]["volume-times-to-end-dist"][0]),
        yerr=np.array(s[dtc]["volume-times-to-end-dist"][1])*1.96,
        marker=marker, ls='-',
        label=f"N={size}")
    volume_scaling_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1])*int(size)**(-v_para/(v_perp*d)),
        np.array(s[dtc]["volume-times-to-end-dist"][0])*int(size)**(-upsilon/(v_perp*d)),
        yerr=np.array(s[dtc]["volume-times-to-end-dist"][1])*1.96*int(size)**(-upsilon/(v_perp*d)),
        marker=marker, ls='-',
        label=f"N={size}")

    mass_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1]),
        np.array(s[dtc]["mass-times-to-end-dist"][0]),
        yerr=np.array(s[dtc]["mass-times-to-end-dist"][1])*1.96,
        marker=marker, ls='-',
        label=f"N={size}")
    mass_scaling_ax.errorbar(
        np.array(s[dtc]["time-bin-edges"][0][:-1])*int(size)**(-v_para/(v_perp*d)),
        np.array(s[dtc]["mass-times-to-end-dist"][0])*int(size)**(-gamma/(v_perp*d)),
        yerr=np.array(s[dtc]["mass-times-to-end-dist"][1])*1.96*int(size)**(-gamma/(v_perp*d)),
        marker=marker, ls='-',
        label=f"N={size}")


  if len(sizes) > 1:
    mk_survival_ax.legend()
  mk_survival_ax.set_xscale('log')
  mk_survival_ax.set_yscale('log')
  mk_survival_ax.set_xlabel("$t$")
  mk_survival_ax.set_ylabel("Survival Probability $\\hat{P}(t)$")

  if len(sizes) > 1:
    mk_survival_scaling_ax.legend()
  mk_survival_scaling_ax.set_xscale('log')
  mk_survival_scaling_ax.set_yscale('log')
  mk_survival_scaling_ax.set_xlabel("$t N^{-\\nu_\\parallel/{\\nu_\\perp d}}$")
  mk_survival_scaling_ax.set_ylabel("$\\hat{P}(t) N^{\\delta/\\nu_\\perp d}$")

  if len(sizes) > 1:
    mass_ax.legend()
  mass_ax.set_xscale('log')
  mass_ax.set_yscale('log')
  mass_ax.set_xlabel("$t$")
  mass_ax.set_ylabel("Mean Component Mass $M(t)$")

  if len(sizes) > 1:
    mass_scaling_ax.legend()
  mass_scaling_ax.set_xscale('log')
  mass_scaling_ax.set_yscale('log')
  mass_scaling_ax.set_xlabel("$t N^{-\\nu_\\parallel/{\\nu_\\perp d}}$")
  mass_scaling_ax.set_ylabel("$M(t) N^{-\\gamma/{\\nu_\\perp d}}$")

  if len(sizes) > 1:
    volume_ax.legend()
  volume_ax.set_xscale('log')
  volume_ax.set_yscale('log')
  volume_ax.set_xlabel("$t$")
  volume_ax.set_ylabel("Mean Spatial Volume $V(t)$")

  if len(sizes) > 1:
    volume_scaling_ax.legend()
  volume_scaling_ax.set_xscale('log')
  volume_scaling_ax.set_yscale('log')
  volume_scaling_ax.set_xlabel("$t N^{-\\nu_\\parallel/{\\nu_\\perp d}}$")
  volume_scaling_ax.set_ylabel("$V(t) N^{-\\upsilon/{\\nu_\\perp d}}$")

  prefix = "figures/single"
  suffix = f"{static_model.value}-{deg_id}-{temporal_model.value}.pdf"

  mk_survival_fig.savefig(
      f"{prefix}/km-survival-prob-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(mk_survival_fig)

  mk_survival_scaling_fig.savefig(
      f"{prefix}/km-survival-scaling-prob-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(mk_survival_scaling_fig)

  mass_fig.savefig(
      f"{prefix}/mass-vs-time-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(mass_fig)

  mass_scaling_fig.savefig(
      f"{prefix}/mass-vs-time-scaling-prob-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(mass_scaling_fig)

  volume_fig.savefig(
      f"{prefix}/volume-vs-time-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(volume_fig)

  volume_scaling_fig.savefig(
      f"{prefix}/volume-vs-time-scaling-prob-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  plt.close(volume_scaling_fig)

def plot_collapse_vs_delta(summaries,
    static_model, temporal_model, prob_dist, deg_id,
    dtcs, beta, v_para, v_perp, d):
  delta = beta/v_para
  beta_p = beta
  z = v_para/v_perp

  theta = d/z - 2*delta
  gamma = v_para + d*v_perp - beta - beta_p
  tau = v_para - beta_p
  nu = d*v_perp- beta_p

  mass_fig,     mass_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  volume_fig,   volume_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  lifetime_fig, lifetime_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))

  mass_ax.set_yscale('log')
  volume_ax.set_yscale('log')
  lifetime_ax.set_yscale('log')

  scaled_mass_fig,      scaled_mass_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  scaled_volume_fig,    scaled_volume_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))
  scaled_lifetime_fig,  scaled_lifetime_ax = plt.subplots(figsize=(3.375*1.5, 2.08*1.5))

  scaled_mass_ax.axvline(0, ls='--', color='grey', label=f"$\\tau=0$")
  scaled_volume_ax.axvline(0, ls='--', color='grey', label=f"$\\tau=0$")
  scaled_lifetime_ax.axvline(0, ls='--', color='grey', label=f"$\\tau=0$")

  # scaled_mass_ax.set_xscale('log')
  scaled_mass_ax.set_yscale('log')
  # scaled_volume_ax.set_xscale('log')
  scaled_volume_ax.set_yscale('log')
  # scaled_lifetime_ax.set_xscale('log')
  scaled_lifetime_ax.set_yscale('log')

  sizes = sorted(summaries, key=int)

  for size, marker in zip(tqdm(sizes), [".", "v", "h", "^", "*", ">", "2", "<", "d"]):
    s = summaries[size][prob_dist.name]

    dt_strings = sorted(s, key=float)
    dts = np.array([float(dt) for dt in dt_strings])
    taus = (dts - dtcs[int(size)])*int(size)**(1/(d*v_perp))

    masses = np.array([s[dt_s]["mean-mass"][0] for dt_s in dt_strings])
    mass_ax.plot(dts, masses,
        label=f"N={size}", marker=marker)
    scaled_mass_ax.plot(taus,
        masses*int(size)**(-gamma/(d*v_perp)),
        label=f"N={size}", marker=marker)


    volumes = np.array([s[dt_s]["mean-volume"][0] for dt_s in dt_strings])
    volume_ax.plot(dts, volumes,
        label=f"N={size}", marker=marker)
    scaled_volume_ax.plot(taus,
        volumes*int(size)**(-nu/(d*v_perp)),
        label=f"N={size}", marker=marker)


    lifetimes = np.array([s[dt_s]["mean-lifetime"][0] for dt_s in dt_strings])
    lifetime_ax.plot(dts, lifetimes,
        label=f"N={size}", marker=marker)
    scaled_lifetime_ax.plot(taus,
        lifetimes*int(size)**(-tau/(d*v_perp)),
        label=f"N={size}", marker=marker)


  mass_ax.set_xlabel("$\\delta t$")
  mass_ax.set_ylabel("Mean Component Mass $M$")
  if len(sizes) > 1:
    mass_ax.legend()

  volume_ax.set_xlabel("$\\delta t$")
  volume_ax.set_ylabel("Mean Spatial Volume $V$")
  if len(sizes) > 1:
    volume_ax.legend()

  lifetime_ax.set_xlabel("$\\delta t$")
  lifetime_ax.set_ylabel("Mean Lifetime $T$")
  if len(sizes) > 1:
    lifetime_ax.legend()

  scaled_mass_ax.set_xlabel(
      "$(\\delta t - \\delta t_c) N^{1/d v_\\perp}$")
  scaled_mass_ax.set_ylabel("$M N^{\\gamma/{d v_\\perp}}$")
  if len(sizes) > 1:
    scaled_mass_ax.legend()
  scaled_mass_ax.set_xlim(right=1)

  scaled_volume_ax.set_xlabel(
      "$(\\delta t - \\delta t_c) N^{1/d v_\\perp}$")
  scaled_volume_ax.set_ylabel("$V N^{\\upsilon/{d v_\\perp}}$")
  if len(sizes) > 1:
    scaled_volume_ax.legend()
  scaled_volume_ax.set_xlim(right=1)

  scaled_lifetime_ax.set_xlabel(
      "$(\\delta t - \\delta t_c) N^{1/d v_\\perp}$")
  scaled_lifetime_ax.set_ylabel("$T N^{\\tau/{d v_\\perp}}$")
  if len(sizes) > 1:
    scaled_lifetime_ax.legend()
  scaled_lifetime_ax.set_xlim(right=1)


  prefix = "figures/single"
  suffix =\
      f"{static_model.value}-{deg_id}-"\
      f"{temporal_model.value}.pdf"

  mass_fig.savefig(f"{prefix}/mean-mass-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  volume_fig.savefig(f"{prefix}/mean-volume-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  lifetime_fig.savefig(f"{prefix}/mean-lifetime-{suffix}",
      bbox_inches='tight', pad_inches=0.1)

  scaled_mass_fig.savefig(f"{prefix}/scaled-mean-mass-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  scaled_volume_fig.savefig(f"{prefix}/scaled-mean-volume-{suffix}",
      bbox_inches='tight', pad_inches=0.1)
  scaled_lifetime_fig.savefig(f"{prefix}/scaled-mean-lifetime-{suffix}",
      bbox_inches='tight', pad_inches=0.1)

  plt.close(mass_fig)
  plt.close(volume_fig)
  plt.close(lifetime_fig)

  plt.close(scaled_mass_fig)
  plt.close(scaled_volume_fig)
  plt.close(scaled_lifetime_fig)


sources = chain(
    product(
      [StaticModel.regular],
      [TemporalModel.poisson],#, TemporalModel.bursty_low, TemporalModel.bursty_medium,
      #  TemporalModel.bursty_high, TemporalModel.bursty_veryhigh, TemporalModel.bursty_veryveryhigh],
      [9]),
    # product(
    #   [StaticModel.heavytail_low, StaticModel.heavytail_medium, StaticModel.heavytail_high,
    #    StaticModel.heavytail_veryhigh, StaticModel.heavytail_veryveryhigh],
    #   [TemporalModel.poisson],
    #   [9]),
    # product(
    #   [StaticModel.grid],
    #   [TemporalModel.poisson],
    #   [1]),
    # product(
    #   [StaticModel.er],
    #   [TemporalModel.poisson, TemporalModel.bursty_low, TemporalModel.bursty_medium,
    #    TemporalModel.bursty_high, TemporalModel.bursty_veryhigh, TemporalModel.bursty_veryveryhigh],
    #   [8])
    )

beta =   {1: 0.276, 2: 0.583, 3: 0.818, 4: 1.0}
v_para = {1: 1.733, 2: 1.287, 3: 1.106, 4: 1.0}
v_perp = {1: 1.096, 2: 0.729, 3: 0.584, 4: 0.5}

for static, temporal, deg_id in sources:
  d = 4

  # for grid model, deg_id represents dimensions of grid
  if static is StaticModel.grid:
    if deg_id < 4:
      d = deg_id
    else:
      d = 4
  else:
    d = 4

  dtcs = {}
  if static is StaticModel.erdos:
    if temporal is TemporalModel.poisson:
      dtcs = {
          32: 0.0930, 64: 0.0890, 128: 0.0860,
          256: 0.0850, 512: 0.0840, 1024: 0.0840,
          2048: 0.0840, 4096: 0.0840, 8192: 0.0840,
          16384: 0.0840, 32768: 0.0840, 65536: 0.0840,
          131072: 0.0840, 262144: 0.0840}
    elif temporal is TemporalModel.bursty_low:
      dtcs = {
          32: 0.0810, 64: 0.0760, 128: 0.0740,
          256: 0.0730, 512: 0.0720, 1024: 0.0710,
          2048: 0.0710, 4096: 0.0710, 8192: 0.0710,
          16384: 0.0710, 32768: 0.0710, 65536: 0.0710,
          131072: 0.0710, 262144: 0.0710}
    else:
      dtcs = {
          32: 0.0940, 64: 0.0910, 128: 0.0890,
          256: 0.0880, 512: 0.0870, 1024: 0.0870,
          2048: 0.0870, 4096: 0.0870, 8192: 0.0870,
          16384: 0.0870, 32768: 0.0870, 65536: 0.0870,
          131072: 0.0870, 262144: 0.0870}
  elif static is StaticModel.regular:
    if temporal is TemporalModel.poisson:
      dtcs = defaultdict(lambda: 0.08802)
    elif temporal is TemporalModel.bursty_low:
      dtcs = defaultdict(lambda: 0.0740)
    else:
      dtcs = defaultdict(lambda: 0.0910)
  elif static is StaticModel.grid and \
          temporal is TemporalModel.poisson and\
          deg_id == 1:
    dtcs = defaultdict(lambda: 0.97700)
  elif static is StaticModel.grid and \
          temporal is TemporalModel.poisson and\
          deg_id == 2:
    dtcs = defaultdict(lambda: 0.28350)
  elif static is StaticModel.grid and \
          temporal is TemporalModel.poisson and\
          deg_id == 3:
    dtcs = defaultdict(lambda: 0.15100)
  elif static is StaticModel.grid and \
          temporal is TemporalModel.poisson and\
          deg_id == 4:
    dtcs = defaultdict(lambda: 0.09300)
  elif static is StaticModel.grid and \
          temporal is TemporalModel.poisson and\
          deg_id == 5:
    dtcs = defaultdict(lambda: 0.08900)
  elif static is StaticModel.heavytail_low and \
          temporal is TemporalModel.poisson:
    dtcs = defaultdict(lambda: 0.00100)
  elif static is StaticModel.heavytail_medium and \
          temporal is TemporalModel.poisson:
    dtcs = defaultdict(lambda: 0.00100)
  elif static is StaticModel.heavytail_high and \
          temporal is TemporalModel.poisson:
    dtcs = defaultdict(lambda: 0.00100)
  elif static is StaticModel.heavytail_veryhigh and \
          temporal is TemporalModel.poisson:
    dtcs = defaultdict(lambda: 0.00150)
  elif static is StaticModel.heavytail_veryveryhigh and \
          temporal is TemporalModel.poisson:
    dtcs = defaultdict(lambda: 0.00350)


  prefix = "networks/aggregated_networks"
  filename = f"{prefix}/{static.name}-{temporal.name}-{deg_id}-single.json"
  print(filename)

  prob_dist = ProbDist.deterministic

  if os.path.isfile(filename):
    with open(filename) as f:
      try:
        summaries = json.load(f)

        plot_collapse_vs_delta(
            summaries,
            static, temporal, prob_dist, deg_id,
            dtcs, beta[d], v_para[d], v_perp[d], d)

        plot_collapse_vs_time(
            summaries,
            static, temporal, prob_dist, deg_id,
            dtcs, beta[d], v_para[d], v_perp[d], d)
      except json.JSONDecodeError:
        continue

class RealNetworks(enum.Enum):
  air = "air"
  transport = "transport"
  twitter = "twitter"
  mobile = "mobile"

class RealTemporal(enum.Enum):
  real = "real"

for real in [RealNetworks.air, RealNetworks.transport, RealNetworks.twitter, RealNetworks.mobile]:
  filename = f"networks/{real.name}/single.json"
  print(filename)
  
  prob_dist = ProbDist.deterministic
  temporal =  RealTemporal.real
  
  d = 4
  if real is RealNetworks.air:
    dtcs = {279: 500} 
  elif real is RealNetworks.transport:
    dtcs = {6986: 600}
  elif real is RealNetworks.twitter:
    dtcs = {17313623: 10}
  elif real is RealNetworks.mobile:
    dtcs = {5193086: 27000}
  
  if os.path.isfile(filename):
    with open(filename) as f:
      try:
        summaries = json.load(f)
  
        plot_collapse_vs_delta(
            summaries,
            real, temporal, prob_dist, deg_id,
            dtcs, beta[d], v_para[d], v_perp[d], d)
  
        plot_collapse_vs_time(
            summaries,
            real, temporal, prob_dist, deg_id,
            dtcs, beta[d], v_para[d], v_perp[d], d)
      except json.JSONDecodeError:
        continue
