import json
import argparse
from itertools import product

import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from network_configs import StaticModel, TemporalModel, ProbDist

import matplotlib 
matplotlib.rc('legend', fontsize='x-small', handlelength=1, markerscale=2)
matplotlib.rc('lines', linewidth=0.75, markersize=2)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

if __name__=="__main__":
  parser = argparse.ArgumentParser(description="Plots quantities of random networks")
  parser.add_argument('static', type=StaticModel)
  parser.add_argument('temporal', type=TemporalModel)
  args = parser.parse_args()

  static = args.static
  temporal = args.temporal
  prob_dist = ProbDist.deterministic

  print("plotting", static, temporal, prob_dist)

  deg_id = 9

  if static is StaticModel.erdos:
    deg_id = 8

  markers = [".", "v", "h", "^", "*", ">", "2", "<", "d", "+", "x"]
  fig, ax = plt.subplots(figsize=(3, 2))

  with open(f"networks/aggregated_networks/{static.value}-{temporal.value}-{deg_id}-degree-dist.json") as f:
    dist_summaries = json.load(f)
    sizes = sorted(dist_summaries, key=int)

    for size_s, marker in zip(sizes, markers):
      s = dist_summaries[size_s][prob_dist.value]
      size = int(size_s)

      dts = []
      counts = {}
      mean_excess = []
      for i, j in product(range(3), repeat=2):
        counts[(i, j)] = []

      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      for dt_s in keys:
        total = 0
        mean_excess_denom = 0
        mean_excess_nom = 0
        for cnt in s[dt_s]["counts"]:
          total += cnt["count"] 
          mean_excess_denom += cnt["count"]*cnt["in"]
        for cnt in s[dt_s]["counts"]:
          counts[(cnt["in"], cnt["out"])].append(cnt["count"]/total)
          mean_excess_nom += cnt["count"]*cnt["in"]*cnt["out"]
        mean_excess.append(mean_excess_nom/mean_excess_denom)
        dts.append(s[dt_s]["dt"][0])

      ax.axhline(0.0, color='grey', ls='--')
      ax.plot(np.array(dts), np.array(mean_excess) - 1, ls='--', marker=marker)
      f = interpolate.UnivariateSpline(mean_excess, dts, ext='raise')
      dtc = f(1.0)
      print(size, dtc)
      delta = 1e-2
      deriv = f.derivative()(1.0)
      ax.set_xlim(dtc-delta, dtc+delta)
      ax.set_ylim(-1.0/deriv*delta*1.05, 1.0/deriv*delta*1.05)
     #  for i, j in product(range(3), repeat=2):
     #    ax.plot(dts, counts[(i, j)])

  ax.set_xlabel("$\\delta t$")
  ax.set_ylabel("$\\tau = \\langle Q_{out} \\rangle-1$")
  fig.savefig(f"figures/degree-dist-{static.value}-{temporal.value}-{deg_id}.pdf",
      bbox_inches='tight', pad_inches=0.1)
