import json
import numpy as np
import matplotlib.pyplot as plt

from network_configs import StaticModel, TemporalModel, ProbDist

import matplotlib 
matplotlib.rc('legend', fontsize='small', handlelength=1, markerscale=2)
matplotlib.rc('lines', linewidth=0.75, markersize=3)
matplotlib.rc('axes', labelpad=2)
matplotlib.rc('axes.spines', top=False, right=False)
matplotlib.rc('xtick.major', pad=2)
matplotlib.rc('xtick.minor', pad=2)
matplotlib.rc('ytick.major', pad=2)
matplotlib.rc('ytick.minor', pad=2)

if __name__=="__main__":
  fig, (ax_tau, ax_mass) = plt.subplots(nrows=2, ncols=1, figsize=(4.0, 5.0), gridspec_kw={'hspace': 0.3})
  
  dts = np.linspace(0, 1.0, num=40)
  lam = 1.0
  colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
  markers = [".", "v", "h", "^", "*", ">", "2", "<", "d", "+", "x"]
  for k, c, m in zip(range(2, 10), colors, markers):
    taus = ((4*k**2 - 6*k + 2)*np.exp(dts*(-k)*lam) + k*np.exp(dts*(1-2*k)*lam)-2*(k-1)**2)/((1-2*k)*k)
    ax_tau.plot(dts, taus, label=f"$k={k}$", color=c, marker=m)
  ax_tau.axhline(y=0.0, ls='--', color='grey')
  
  ax_tau.text(-0.15, 0.95, f"(a)", transform=ax_tau.transAxes)
  ax_tau.set_xlabel("$\\delta t$")
  ax_tau.set_ylabel("$\\tau$")

  ax_tau.legend(ncol=2)

  k = 9
  dtc = 0.08559
  dtc_real = 0.08808
  ax_mass.axvline(x=dtc, ls='--', color='grey', label=f"$\\delta t_c^{{theoretical}} = {dtc:.5f}$")
  ax_mass.axvline(x=dtc_real, ls=':', color='grey', label=f"$\\delta t_c^{{empirical}} = {dtc_real:.5f}$")

  dts = dtc - np.geomspace(2e-4, dtc, num=35)
  ax_mass.plot(dts, 0.126*(dtc-dts)**(-1), marker=markers[1], label="$c (\\delta t_c^{theoretical} - \\delta t)^{-1}$")

  dts_real = dtc_real - np.geomspace(2e-4, dtc_real, num=35)
  ax_mass.plot(dts_real, 0.126*(dtc_real-dts_real)**(-1), marker=markers[2], label="$c (\\delta t_c^{empirical} - \\delta t)^{-1}$")

  dts = dtc - np.geomspace(2e-4, dtc, num=40)
  lam = 1.0
  taus = ((4*k**2 - 6*k + 2)*np.exp(dts*(-k)*lam) + k*np.exp(dts*(1-2*k)*lam)-2*(k-1)**2)/((1-2*k)*k)
  masses = (-taus)**(-1)
  ax_mass.plot(dts[masses > 0], masses[masses > 0], marker=markers[0], label="$(-\\tau)^{-1}$")

  with open("networks/aggregated_networks/regular-poisson-9-single.json") as f:
    single_summaries = json.load(f)
    sizes = sorted(single_summaries, key=int)
    for size_s, c, m in zip(sizes[-1:], colors[3:], markers[3:]):
      s = single_summaries[size_s]["deterministic"]
      keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
      keys = [k for k in keys if s[k]["dt"][0] <= dtc_real]
      dts = np.array([s[k]["dt"][0] for k in keys])
      masses = np.array([s[k]["mean-mass"][0] for k in keys])
      ax_mass.plot(dts, masses, marker=m, ls='--', label="Empirical $M(\\delta t)$")

  # with open("networks/aggregated_networks/nsi_regular-poisson-9-single.json") as f:
  #   single_summaries = json.load(f)
  #   sizes = sorted(single_summaries, key=int)
  #   for size_s, c, m in zip(sizes[-1:], colors[3:], markers[3:]):
  #     s = single_summaries[size_s]["deterministic"]
  #     keys = sorted(s.keys(), key=lambda k: s[k]["dt"][0])
  #     keys = [k for k in keys if s[k]["dt"][0] <= dtc_real]
  #     dts = np.array([s[k]["dt"][0] for k in keys])
  #     masses = np.array([s[k]["mean-mass"][0] for k in keys])
  #     ax_mass.plot(dts, masses, marker=m, ls='--', label="Empirical (NSI) $M(\\delta t)$")

  ax_mass.text(-0.15, 0.95, f"(b)", transform=ax_mass.transAxes)
  ax_mass.set_xlabel("$\\delta t$")
  ax_mass.set_ylabel("$M(\\delta t)$")
  ax_mass.set_yscale("log")
  ax_mass.legend()

  fig.savefig(f"figures/theoritical-tau-mass.pdf",
      bbox_inches='tight', pad_inches=0.2)
  
