#!/bin/bash

set -e

module purge

if [[ -z $static ]]; then  # temporal network generation
  exit 1
elif [[ -z $temporal ]]; then
  exit 1
elif [[ -z $nodes ]]; then
  exit 1
elif [[ -z $max_time ]]; then
  exit 1
elif [[ -z $k ]]; then
  exit 1
elif [[ -z $dtc ]]; then  # dt list generation
  exit 1
elif [[ -z $precision ]]; then
  exit 1
elif [[ -z $perjob ]]; then  # parallelisation/realisation
  exit 1
elif [[ -z $SLURM_PROCID ]]; then
  exit 1
elif [[ -z $SLURM_CPUS_PER_TASK ]]; then
  exit 1
fi


static_dir=${static};
temporal_dir=${temporal};

case $static in
  "grid")
    if [[ -z $dims ]]; then  # dimensions for grid
      exit 1
    fi
    static_dir=${static}_${dims};
  ;;
esac

case $temporal in
  "bursty")
    if [[ -z $exponent ]]; then  # exponent for bursty
      exit 1
    fi
    temporal_dir=${temporal}_${exponent};
  ;;
  "self-exciting")
    if [[ -z $mu  || -z $alpha || -z $theta ]]; then
      exit 1
    fi
    temporal_dir=${temporal}_${mu}_${alpha}_${theta};
  ;;
esac

tmpdir=$(mktemp -d --tmpdir=/dev/shm)
task_id=$(( (SLURM_PROCID + 1) * perjob ))
prefix=networks/random_large_networks/${static_dir}/${temporal_dir}/$nodes/$k;

function cleanup() {
  rm -rf $tmpir;
}


mkdir -p $tmpdir
trap cleanup EXIT ERR INT TERM
mkdir -p $prefix

seed_seq=$(seq $task_id $(( $task_id + $perjob - 1)))

# Generate temporal networks
if [[ $static == "erdos" && $temporal == "poisson" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "erdos" && $temporal == "bursty" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal --exponent=$exponent \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "erdos" && $temporal == "self-exciting" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal \
    --background-intensity=$mu --infectivity=$alpha --exponential-delay-density-rate=$theta \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "regular" && $temporal == "poisson" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "regular" && $temporal == "bursty" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal --exponent=$exponent \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "regular" && $temporal == "self-exciting" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --max-time $max_time --mean-dt 1 \
    --static-model $static --temporal-model $temporal \
    --background-intensity=$mu --infectivity=$alpha --exponential-delay-density-rate=$theta \
    --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
elif [[ $static == "grid" && $temporal == "poisson" ]]; then
  parallel -j $SLURM_CPUS_PER_TASK --header : --shuf -v \
    ./random_network --seed {seed} --nodes $nodes --dims $dims \
    --max-time $max_time --mean-dt 1 --static-model $static \
    --temporal-model $temporal --average-degree $k '>' $tmpdir/{seed}.events \
    ::: seed $seed_seq
else
  exit 1
fi


base=$(python3 get_precision_digits.py $dtc $precision);
tmpfile_prefix="$tmpdir/s-${task_id}-base-${base}";

# Generate dt list

python3 generate_dts.py $dtc $precision > $tmpdir/dts.csv

# Run the degree analysis

parallel -j $SLURM_CPUS_PER_TASK --trim lr --colsep ',' --header : --shuf \
  ./network_degree_dist --network $tmpdir/{seed}.events --seed {seed} \
    --sample-size 100000 --degree-dists-file /dev/stdout --dt {dt} "||" true \
    ::: seed $seed_seq :::: $tmpdir/dts.csv \
    > ${tmpfile_prefix}-degree-dist.jsonl;

# Run the single source analysis

parallel -j $SLURM_CPUS_PER_TASK --trim lr --colsep ',' --header : --shuf \
  ./random_large_single \
    --network $tmpdir/{seed}.events \
    --dt {dt} --seed 1 --time-bins 48 --time-bins-min 0.5 --time-bins-max $max_time \
    --window-min 0 --window-max $max_time --summary /dev/stdout \
    ::: seed $seed_seq :::: $tmpdir/dts.csv \
    > ${tmpfile_prefix}-single.jsonl;

# Run the homogenous analysis

parallel -j $SLURM_CPUS_PER_TASK --trim lr --colsep ',' --header : --shuf \
  ./random_large_homogeneous_alternative \
    --network $tmpdir/{seed}.events \
    --dt {dt} --seed 1 --time-bins 48 --time-bins-min 0.5 --time-bins-max $max_time \
    --window-min 0 --window-max $max_time --summary /dev/stdout \
    ::: seed $seed_seq :::: $tmpdir/dts.csv \
    > ${tmpfile_prefix}-homogeneous.jsonl;

mv $tmpdir/*.jsonl $prefix/
