#include <vector>
#include <functional>
#include <algorithm>
#include <cmath>

namespace binned_math {
  template<class KeyT>
  std::vector<KeyT> linear_bins(KeyT min, KeyT max, size_t num) {
    std::vector<KeyT> bin_edges = {min};
    KeyT step = static_cast<KeyT>(
        static_cast<double>(max - min)/static_cast<double>(num));
    if (std::is_integral<KeyT>::value)
      step = std::max(static_cast<KeyT>(1), step);

    bin_edges.reserve(num + 1);

    for (size_t i = 1; i < num; i++)
      bin_edges.push_back(bin_edges.back() + step);

    bin_edges.push_back(max);

    return bin_edges;
  }

  template<class KeyT>
  std::vector<KeyT> logarithmic_bins(KeyT min, KeyT max, size_t num) {
    std::vector<KeyT> bin_edges = {min};
    double current_log = std::log(min);
    double exp_step = (std::log(max) - std::log(min))/static_cast<double>(num);
    bin_edges.reserve(num + 1);

    for (size_t i = 1; i < num; i++) {
      if (std::is_integral<KeyT>::value &&
          std::exp(current_log + exp_step) - std::exp(current_log) < 1) {
        bin_edges.push_back(bin_edges.back() + static_cast<KeyT>(1));
        current_log = std::log(bin_edges.back());
      } else {
        current_log += exp_step;
        bin_edges.push_back(static_cast<KeyT>(std::exp(current_log)));
      }
    }

    bin_edges.push_back(max);

    return bin_edges;
  }

  // bins are [b0, b1), [b1, b2) ..., [bn, bn+1]
  template<class KeyT>
  size_t bin_number(
      const KeyT& key,
      const std::vector<KeyT>& bin_edges) {
    if (key < bin_edges.front())
      return bin_edges.size();
    else if (key > bin_edges.back())
      return bin_edges.size();
    else if (key == bin_edges.back())
      return bin_edges.size() - 2ul;
    else
      return std::distance(bin_edges.begin(),
          std::upper_bound(
            bin_edges.begin(), bin_edges.end(), key)) - 1ul;
  }

  template<class KeyT, class ValueT, class StatisticT>
  std::vector<StatisticT> binned_statistic(
      const std::vector<KeyT>& keys,
      const std::vector<ValueT>& values,
      const std::vector<KeyT>& bin_edges,
      std::function<StatisticT(const std::vector<ValueT>&)> statistic) {
    if (!std::is_sorted(bin_edges.begin(), bin_edges.end()))
      throw std::invalid_argument(
          "bin_edges should be sorted");
    if (bin_edges.size() < 2)
      throw std::invalid_argument(
          "bin_edges should contain at least two items");
    if (keys.size() != values.size())
      throw std::invalid_argument(
          "keys and values should be of the same size");

    // TODO: Use a vector
    std::unordered_map<size_t, std::vector<ValueT>> val_bins;
    for (size_t i = 0; i < keys.size(); i++)
      val_bins[bin_number(keys[i], bin_edges)].push_back(values[i]);

    std::vector<StatisticT> results;
    results.reserve(bin_edges.size());
    for (size_t i = 0; i < bin_edges.size() - 1; i++)
      results.push_back(statistic(val_bins[i]));

    return results;
  }

  template<class ValueT>
  ValueT sum(const std::vector<ValueT>& values) {
    return std::accumulate(values.begin(), values.end(), ValueT{});
  }

  template<class ValueT>
  size_t count(const std::vector<ValueT>& values) {
    return values.size();
  }

  template<class ValueT>
  double mean(const std::vector<ValueT>& values) {
    if (values.size() == 0)
      return std::nan("");

    // cast values to double to prevent overflow of integral types in sum()
    std::vector<double> double_vals(values.begin(), values.end());

    return sum(double_vals)/static_cast<double>(double_vals.size());
  }
}
