#include <string>
#include <fstream>
#include <unordered_map>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop
#include <nlohmann/json.hpp>

#include <dag/dag.hpp>
#include <hyperloglog.hpp>

enum class prob_dist_types { deterministic, exponential };

template <class EdgeT>
struct opts {
  typename EdgeT::TimeType dt;
  size_t seed;
  size_t sample_size;
  size_t temporal_reserve;

  std::string network_filename;
  std::string degree_dists_filename;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("network_degree_dist",
        "estimates joint in- and out- degree probability distribution of");
    option_def.add_options()
      ("seed", "seed for random sampling",
      cxxopts::value<size_t>())
      ("dt", "delta-t value for this network",
      cxxopts::value<typename EdgeT::TimeType>())
      ("sample-size", "number of samples",
      cxxopts::value<size_t>()->default_value("1"))
      ("h,help", "Print help")
      ;

    option_def.add_options("Event List File")
      ("temporal-reserve", "estimated size of temporal network",
      cxxopts::value<size_t>()->default_value("0"))
      ("network", "network file",
      cxxopts::value<std::string>());

    option_def.add_options("Output")
      ("degree-dists-file", "file to store degree distributions",
      cxxopts::value<std::string>())
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("dt") == 0) {
      std::cerr << "ERROR: needs a dt argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("network") == 0) {
      std::cerr << "ERROR: needs a network argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("degree-dists-file") == 0) {
      std::cerr << "ERROR: needs a out-component-growth file" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    }


    seed = options["seed"].as<size_t>();
    network_filename = options["network"].as<std::string>();
    temporal_reserve = options["temporal-reserve"].as<size_t>();
    sample_size = options["sample-size"].as<size_t>();
    dt = options["dt"].as<typename EdgeT::TimeType>();
    degree_dists_filename = options["degree-dists-file"].as<std::string>();
  }
};


struct hash_pair {
  template <class T1, class T2>
  size_t operator()(const std::pair<T1, T2>& p) const {
    return dag::utils::combine_hash(std::hash<T1>{}(p.first), p.second);
  }
};

template <class EdgeT>
std::vector<EdgeT>
event_list(
    std::string net_filename,
    size_t temporal_reserve) {

  std::vector<EdgeT> topo;
  if (temporal_reserve > 0)
    topo.reserve(temporal_reserve);

  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}


using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;

using ProbType = dag::adjacency_prob::deterministic<EdgeType>;

template <class EdgeT>
size_t minimum_in_degree(const EdgeT& root, const std::vector<EdgeT>& preds) {
  for (auto it = preds.rbegin(); it < preds.rend(); it++)
    if (it->mutated_verts() == root.mutated_verts() &&
        it->mutator_verts() == root.mutator_verts())
      return std::max(
          static_cast<size_t>(std::distance(preds.rbegin(), it)), 1ul);

  return preds.size();
}

template <class EdgeT>
size_t minimum_out_degree(const EdgeT& root, const std::vector<EdgeT>& succs) {
  for (auto it = succs.begin(); it < succs.end(); it++)
    if (it->mutated_verts() == root.mutated_verts() &&
        it->mutator_verts() == root.mutator_verts())
      return std::max(
          static_cast<size_t>(std::distance(succs.begin(), it)), 1ul);

  return succs.size();
}

template <class EdgeT>
void tests() {
  dag::network<EdgeT> net1({
    {1, 2, 0.0},
    {1, 2, 1.0},
    {2, 3, 2.0},
    {0, 1, 3.0}});
  dag::network<EdgeT> net1r({
    {1, 2, 3.0},
    {1, 2, 2.0},
    {2, 3, 1.0},
    {0, 1, 0.0}});
  std::unordered_map<double, size_t> correct1 = {
    {0.5, 0},
    {1.5, 1},
    {2.5, 1},
    {3.5, 1}};

  dag::network<EdgeT> net2({
    {1, 2, 0.0},
    {2, 3, 1.0},
    {1, 2, 2.0},
    {0, 1, 3.0}});
  dag::network<EdgeT> net2r({
    {1, 2, 3.0},
    {2, 3, 2.0},
    {1, 2, 1.0},
    {0, 1, 0.0}});
  std::unordered_map<double, size_t> correct2 = {
    {0.5, 0},
    {1.5, 1},
    {2.5, 1},
    {3.5, 1}};

  dag::network<EdgeT> net3({
    {1, 2, 0.0},
    {2, 3, 1.0},
    {0, 1, 2.0},
    {1, 2, 3.0}});
  dag::network<EdgeT> net3r({
    {1, 2, 3.0},
    {2, 3, 2.0},
    {0, 1, 1.0},
    {1, 2, 0.0}});
  std::unordered_map<double, size_t> correct3 = {
    {0.5, 0},
    {1.5, 1},
    {2.5, 2},
    {3.5, 2}};

  EdgeT root = {1, 2, 0.0};
  EdgeT rootr = {1, 2, 3.0};

  for (double dt: std::vector<double>({0.5, 1.5, 2.5, 3.5})) {
    std::cerr << "dt: " << dt << std::endl;
    ProbType prob(dt);

    dag::implicit_event_graph<EdgeType, ProbType> eg1(net1, prob, 0ul);
    dag::implicit_event_graph<EdgeType, ProbType> eg1r(net1r, prob, 0ul);
    std::cerr << "net1: " <<
      minimum_out_degree(root, eg1.successors(root, true)) <<
      " correct1: " << correct1[dt] << std::endl;
    assert(minimum_out_degree(root,
        eg1.successors(root, true)) == correct1[dt]);
    std::cerr << "net1r: " <<
      minimum_in_degree(rootr, eg1r.predecessors(rootr, true)) <<
      " correct1: " << correct1[dt] << std::endl;
    assert(minimum_in_degree(rootr,
        eg1r.predecessors(rootr, true)) == correct1[dt]);

    dag::implicit_event_graph<EdgeType, ProbType> eg2(net2, prob, 0ul);
    dag::implicit_event_graph<EdgeType, ProbType> eg2r(net2r, prob, 0ul);
    std::cerr << "net2: " <<
      minimum_out_degree(root, eg2.successors(root, true)) <<
      " correct2: " << correct2[dt] << std::endl;
    assert(minimum_out_degree(root,
        eg2.successors(root, true)) == correct2[dt]);
    std::cerr << "ne21r: " <<
      minimum_in_degree(rootr, eg2r.predecessors(rootr, true)) <<
      " correct1: " << correct2[dt] << std::endl;
    assert(minimum_in_degree(rootr,
        eg2r.predecessors(rootr, true)) == correct2[dt]);

    dag::implicit_event_graph<EdgeType, ProbType> eg3(net3, prob, 0ul);
    dag::implicit_event_graph<EdgeType, ProbType> eg3r(net3r, prob, 0ul);
    std::cerr << "net3: " <<
      minimum_out_degree(root, eg3.successors(root, true)) <<
      " correct3: " << correct3[dt] << std::endl;
    assert(minimum_out_degree(root,
        eg3.successors(root, true)) == correct3[dt]);
    std::cerr << "net3r: " <<
      minimum_in_degree(rootr, eg3r.predecessors(rootr, true)) <<
      " correct1: " << correct3[dt] << std::endl;
    assert(minimum_in_degree(rootr,
        eg3r.predecessors(rootr, true)) == correct3[dt]);
  }
}

int main(int argc, char* argv[]) {
  opts<EdgeType> options(argc, argv);
  /* tests<EdgeType>(); */

  ProbType prob(options.dt);
  size_t eg_seed = 0ul;

  dag::network<EdgeType> network(
        event_list<EdgeType>(options.network_filename, options.temporal_reserve));

  dag::implicit_event_graph<EdgeType, ProbType> eg(network, prob, eg_seed);

  std::vector<EdgeType> event_samples;
  std::mt19937 gen(options.seed);
  std::sample(
      eg.events_cause().begin(),
      eg.events_cause().end(),
      std::back_inserter(event_samples),
      options.sample_size,
      gen
      );

  std::unordered_map<std::pair<size_t, size_t>, size_t, hash_pair> counts;

  for (auto&& root: event_samples) {
    if (root.cause_time() > eg.time_window().first + options.dt &&
        root.effect_time() + options.dt < eg.time_window().second) {
      size_t in_degree = minimum_in_degree(root, eg.predecessors(root, true));
      size_t out_degree = minimum_out_degree(root, eg.successors(root, true));
      if (in_degree > 2 || out_degree > 2)
        throw(std::logic_error("in-degree or out-degree larger than 2"));
      counts[std::make_pair(in_degree, out_degree)] += 1;
    }
  }

  nlohmann::json summary = {
    {"dt", options.dt},
    {"prob_dist_type", "deterministic"},
    {"seed", options.seed},
    {"counts", {}},
    {"network", {
      {"filename", options.network_filename},
      {"temporal-edges", network.edges_cause().size()},
      {"temporal-nodes", network.vertices().size()},
    }}
  };

  for (size_t in = 0; in < 3; in++)
    for (size_t out = 0; out < 3; out++)
      summary["counts"].push_back(nlohmann::json::object({
          {"in", in}, {"out", out},
          {"count", counts[std::make_pair(in, out)]}}));

  std::ofstream degree_dists_file(options.degree_dists_filename);
  degree_dists_file << summary.dump() << std::endl;
}
