#include <string>
#include <fstream>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop
#include <nlohmann/json.hpp>

#include <dag/dag.hpp>
#include <hyperloglog.hpp>

#include "binned_math.tpp"

enum class prob_dist_types { deterministic, exponential };

template <class EdgeT>
struct opts {
  typename EdgeT::TimeType dt;
  typename EdgeT::TimeType time_bins_min;
  typename EdgeT::TimeType time_bins_max;
  typename EdgeT::TimeType window_min;
  typename EdgeT::TimeType window_max;
  size_t time_bins;
  size_t seed;
  size_t temporal_reserve;
  double field;

  std::string network_filename;
  std::string summary_filename;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("network_homogeneous",
        "Calculates properties of homogeneous spreading on temporal network");
    option_def.add_options()
      ("seed", "seed for cardinality estimators",
      cxxopts::value<size_t>())
      ("dt", "delta-t value for this network",
      cxxopts::value<typename EdgeT::TimeType>())
      ("field", "external field value (rate of spontaneous particle generation)",
      cxxopts::value<double>()->default_value("0.0"))
      ("h,help", "Print help")
      ;

    option_def.add_options("Event List File")
      ("temporal-reserve", "estimated size of temporal network",
      cxxopts::value<size_t>()->default_value("0"))
      ("network", "network file",
      cxxopts::value<std::string>());

    option_def.add_options("Output")
      ("summary", "file to store summary statistics of the network",
      cxxopts::value<std::string>())
      ("window-min", "start of measurement",
      cxxopts::value<typename EdgeT::TimeType>())
      ("window-max", "end of measurement",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins-min", "first time bin edge",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins-max", "last time bin edge",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins", "number of time bins",
      cxxopts::value<size_t>())
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("network") == 0) {
      std::cerr << "ERROR: needs a network argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("summary") == 0) {
      std::cerr << "ERROR: needs a summary file" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("time-bins") == 0) {
      std::cerr << "ERROR: needs a time-bins value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("time-bins-min") == 0) {
      std::cerr << "ERROR: needs a time-bin-min value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("time-bins-max") == 0) {
      std::cerr << "ERROR: needs a time-bin-max value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("window-min") == 0) {
      std::cerr << "ERROR: needs a window-min value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("window-max") == 0) {
      std::cerr << "ERROR: needs a window-max value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    }

    seed = options["seed"].as<size_t>();
    temporal_reserve = options["temporal-reserve"].as<size_t>();
    dt = options["dt"].as<typename EdgeT::TimeType>();
    field = options["field"].as<double>();
    time_bins = options["time-bins"].as<size_t>();
    time_bins_min = options["time-bins-min"].as<typename EdgeT::TimeType>();
    time_bins_max = options["time-bins-max"].as<typename EdgeT::TimeType>();
    window_min = options["window-min"].as<typename EdgeT::TimeType>();
    window_max = options["window-max"].as<typename EdgeT::TimeType>();

    network_filename = options["network"].as<std::string>();
    summary_filename = options["summary"].as<std::string>();
  }
};

template <class EdgeT>
std::vector<EdgeT>
event_list(
    std::string net_filename,
    size_t temporal_reserve) {
  std::vector<EdgeT> topo;
  if (temporal_reserve > 0)
    topo.reserve(temporal_reserve);

  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}

#if defined(NETWORK_MODE_MOBILE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_TRANSPORT)
  using EdgeType = dag::directed_delayed_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_LARGE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_RANDOM_LARGE)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#else
  #error No network mode defined. should define one of the following:\
    `NETWORK_MODE_MOBILE`, `NETWORK_MODE_TRANSPORT`, `NETWORK_MODE_RANDOM`
#endif

using ProbType = dag::adjacency_prob::deterministic<EdgeType>;
template <class T>
using HLLEstimator = dag::hll_estimator<T, HllT>;

template <class T>
using HLLEstimatorRO = dag::hll_estimator_readonly<T, HllT>;
using CompType = dag::temporal_component<EdgeType, HLLEstimatorRO>;

int main(int argc, char* argv[]) {
  opts<EdgeType> options(argc, argv);

  // TODO: add exponential prob_dist_type

  nlohmann::json summary = {
    {"seed", options.seed},
    {"field", options.field},
    {"dt", options.dt},
    {"window", {options.window_min, options.window_max}},
    {"prob_dist_type", "deterministic"},
  };

  auto events = event_list<EdgeType>(
      options.network_filename, options.temporal_reserve);

  typedef std::conditional_t<
    dag::is_undirected_v<EdgeType>,
    dag::undirected_edge<typename EdgeType::VertexType>,
    dag::directed_edge<typename EdgeType::VertexType>> aggregate_edge_type;

  std::vector<aggregate_edge_type> aggregate;
  for (auto&& e: events)
    aggregate.push_back(e.static_projection());

  // TODO handle case of integer TimeType
  std::exponential_distribution<typename EdgeType::TimeType> dist(
      static_cast<typename EdgeType::TimeType>(options.field));
  dag::network<aggregate_edge_type> aggregate_network(aggregate);

  std::vector<EdgeType> spontaneous_particles = dag::random_events<EdgeType>(
      aggregate_network,
      options.window_max - options.window_min,
      dist, dist, 0ul);

  dag::network<EdgeType> network(events);

  summary["network"] = {
    {"filename", options.network_filename},
    {"temporal-edges", network.edges_cause().size()},
    {"temporal-nodes", network.vertices().size()},
    {"spontaneous-particles", spontaneous_particles.size()}
  };


  std::vector<typename EdgeType::TimeType> time_bins =
    binned_math::logarithmic_bins(
      options.time_bins_min,
      options.time_bins_max,
      options.time_bins);

  summary["time-bin-edges"] = time_bins;

  size_t eg_seed = 0ul;
  ProbType prob(options.dt);
  dag::implicit_event_graph<EdgeType, ProbType> eg(network, prob, eg_seed);

  auto in_comps = dag::in_components<
    EdgeType, ProbType, HLLEstimator, HLLEstimatorRO>(eg, options.seed, false);


  std::vector<EdgeType> spontaneous_occupied;
  for (auto&& root: spontaneous_particles) {
    auto comp_exact = dag::out_component<
      EdgeType, ProbType, dag::exact_estimator>(
          eg, root, 0ul, 0ul, 0ul);


    auto new_it = spontaneous_occupied.insert(spontaneous_occupied.end(),
        comp_exact.edge_set().set().begin(),
        comp_exact.edge_set().set().end());

    std::sort(new_it, spontaneous_occupied.end());

    std::inplace_merge(spontaneous_occupied.begin(),
        new_it, spontaneous_occupied.end());
    spontaneous_occupied.erase(
        std::unique(spontaneous_occupied.begin(), spontaneous_occupied.end()),
        spontaneous_occupied.end());
  }

  std::vector<EdgeType> all_events(
      network.edges_cause().begin(),
      network.edges_cause().end());
  all_events.insert(all_events.end(),
      spontaneous_particles.begin(),
      spontaneous_particles.end());

  std::vector<typename EdgeType::TimeType> event_times;
  std::transform(
      all_events.begin(), all_events.end(),
      std::back_inserter(event_times),
      [&options](const EdgeType& e) {
        return e.cause_time() - options.window_min;
      });
  std::vector<typename EdgeType::TimeType> event_times_to_end;
  std::transform(
      all_events.begin(), all_events.end(),
      std::back_inserter(event_times_to_end),
      [&options](const EdgeType& e) {
        return options.window_max - e.cause_time();
      });

  std::vector<EdgeType> occupied_events;
  for (auto [e, ic]: in_comps)
    if (ic.cause_lifetime().first - options.window_min <= options.dt)
      occupied_events.push_back(e);
  occupied_events.insert(occupied_events.end(),
      spontaneous_occupied.begin(),
      spontaneous_occupied.end());
  std::sort(occupied_events.begin(), occupied_events.end());
  occupied_events.erase(
      std::unique(occupied_events.begin(), occupied_events.end()),
      occupied_events.end());

  std::vector<typename EdgeType::TimeType> occupied_event_times;
  std::transform(
      occupied_events.begin(), occupied_events.end(),
      std::back_inserter(occupied_event_times),
      [&options](const EdgeType& e) {
        return e.cause_time() - options.window_min;
      });
  std::vector<typename EdgeType::TimeType> occupied_event_times_to_end;
  std::transform(
      occupied_events.begin(), occupied_events.end(),
      std::back_inserter(occupied_event_times_to_end),
      [&options](const EdgeType& e) {
        return options.window_max - e.cause_time();
      });

  summary["event-times-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
      event_times, event_times, time_bins,
      binned_math::count<typename EdgeType::TimeType>);
  summary["event-times-to-end-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
      event_times_to_end, event_times_to_end, time_bins,
      binned_math::count<typename EdgeType::TimeType>);
  summary["occupied-event-times-dist"] = binned_math::binned_statistic<
    typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
      occupied_event_times, occupied_event_times, time_bins,
      binned_math::count<typename EdgeType::TimeType>);
  summary["occupied-event-times-to-end-dist"] = binned_math::binned_statistic<
    typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
      occupied_event_times_to_end, occupied_event_times_to_end, time_bins,
      binned_math::count<typename EdgeType::TimeType>);

  std::ofstream summary_file(options.summary_filename);
  summary_file << summary.dump() << std::endl;
}
