#include <string>
#include <fstream>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop
#include <nlohmann/json.hpp>

#include <dag/dag.hpp>
#include <hyperloglog.hpp>

#include "binned_math.tpp"

enum class prob_dist_types { deterministic, exponential };

template <class EdgeT>
struct opts {
  typename EdgeT::TimeType dt;

  typename EdgeT::TimeType time_bins_min;

  bool provided_time_bins_max;
  typename EdgeT::TimeType time_bins_max;
  bool provided_window_min;
  typename EdgeT::TimeType window_min;
  bool provided_window_max;
  typename EdgeT::TimeType window_max;

  size_t time_bins;

  size_t seed;
  size_t temporal_reserve;
  double field;

  std::string network_filename;
  std::string summary_filename;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("network_homogeneous",
        "Calculates properties of homogeneous spreading on temporal network");
    option_def.add_options()
      ("seed", "seed for cardinality estimators",
      cxxopts::value<size_t>())
      ("dt", "delta-t value for this network",
      cxxopts::value<typename EdgeT::TimeType>())
      ("h,help", "Print help")
      ;

    option_def.add_options("Event List File")
      ("temporal-reserve", "estimated size of temporal network",
      cxxopts::value<size_t>()->default_value("0"))
      ("network", "network file",
      cxxopts::value<std::string>());

    option_def.add_options("Output")
      ("summary", "file to store summary statistics of the network",
      cxxopts::value<std::string>())
      ("window-min", "start of measurement",
      cxxopts::value<typename EdgeT::TimeType>())
      ("window-max", "end of measurement",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins-min", "first time bin edge",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins-max", "last time bin edge",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins", "number of time bins",
      cxxopts::value<size_t>())
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("network") == 0) {
      std::cerr << "ERROR: needs a network argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("summary") == 0) {
      std::cerr << "ERROR: needs a summary file" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("time-bins") == 0) {
      std::cerr << "ERROR: needs a time-bins value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("time-bins-min") == 0) {
      std::cerr << "ERROR: needs a time-bin-min value" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    }

    if (options.count("window-min") == 0) {
      provided_window_min = false;
    } else {
      provided_window_min = true;
      window_min = options["window-min"].as<typename EdgeT::TimeType>();
    }

    if (options.count("window-max") == 0) {
      provided_window_max = false;
    } else {
      provided_window_max = true;
      window_max = options["window-max"].as<typename EdgeT::TimeType>();
    }

    if (options.count("time-bins-max") == 0) {
      provided_time_bins_max = false;
    } else {
      provided_time_bins_max = true;
      time_bins_max = options["time-bins-max"].as<typename EdgeT::TimeType>();
    }

    seed = options["seed"].as<size_t>();
    temporal_reserve = options["temporal-reserve"].as<size_t>();
    dt = options["dt"].as<typename EdgeT::TimeType>();
    time_bins = options["time-bins"].as<size_t>();
    time_bins_min = options["time-bins-min"].as<typename EdgeT::TimeType>();

    network_filename = options["network"].as<std::string>();
    summary_filename = options["summary"].as<std::string>();
  }

  void update_time_window(std::pair<typename EdgeT::TimeType, typename EdgeT::TimeType> time_window) {
    if (!provided_window_min)
      window_min = time_window.first;

    if (!provided_window_max)
      window_max = time_window.second;

    if (!provided_time_bins_max)
      time_bins_max = time_window.second - time_window.first;
  }
};

template <class EdgeT>
std::vector<EdgeT>
event_list(
    std::string net_filename,
    size_t temporal_reserve) {
  std::vector<EdgeT> topo;
  if (temporal_reserve > 0)
    topo.reserve(temporal_reserve);

  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}

#if defined(NETWORK_MODE_MOBILE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_TRANSPORT)
  using EdgeType = dag::directed_delayed_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_LARGE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED_LARGE)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#else
  #error No network mode defined. should define one of the following:\
    `NETWORK_MODE_MOBILE`, `NETWORK_MODE_TRANSPORT`, `NETWORK_MODE_RANDOM`
#endif

using ProbType = dag::adjacency_prob::deterministic<EdgeType>;
template <class T>
using HLLEstimator = dag::hll_estimator<T, HllT>;

template <class T>
using HLLEstimatorRO = dag::hll_estimator_readonly<T, HllT>;
using CompType = dag::temporal_component<EdgeType, HLLEstimatorRO>;

int main(int argc, char* argv[]) {
  opts<EdgeType> options(argc, argv);

  // TODO: add exponential prob_dist_type
  auto events = event_list<EdgeType>(
      options.network_filename, options.temporal_reserve);

  dag::network<EdgeType> network(events);

  size_t eg_seed = 0ul;
  ProbType prob(options.dt);
  dag::implicit_event_graph<EdgeType, ProbType> eg(network, prob, eg_seed);

  options.update_time_window(eg.time_window());

  nlohmann::json summary = {
    {"seed", options.seed},
    {"dt", options.dt},
    {"window", {options.window_min, options.window_max}},
    {"prob_dist_type", "deterministic"},
  };

  summary["network"] = {
    {"filename", options.network_filename},
    {"temporal-edges", network.edges_cause().size()},
    {"temporal-nodes", network.vertices().size()},
  };

  std::vector<typename EdgeType::TimeType> time_bins =
    binned_math::logarithmic_bins(
      options.time_bins_min,
      options.time_bins_max,
      options.time_bins);

  summary["time-bin-edges"] = time_bins;

  auto in_comps = dag::in_components<
    EdgeType, ProbType, HLLEstimator, HLLEstimatorRO>(eg, options.seed, false);


  std::vector<typename EdgeType::TimeType> event_times;
  std::transform(
      in_comps.begin(), in_comps.end(),
      std::back_inserter(event_times),
      [&options](const std::pair<EdgeType, CompType>& ic) {
        return ic.first.cause_time() - options.window_min;
      });
  std::vector<typename EdgeType::TimeType> event_times_to_end;
  std::transform(
      in_comps.begin(), in_comps.end(),
      std::back_inserter(event_times_to_end),
      [&options](const std::pair<EdgeType, CompType>& ic) {
        return options.window_max - ic.first.cause_time();
      });

  std::vector<double> homogeneous_occupied_p;
  std::transform(
      in_comps.begin(), in_comps.end(),
      std::back_inserter(homogeneous_occupied_p),
      [&options](const std::pair<EdgeType, CompType>& ic) {
        if (ic.second.cause_lifetime().first - options.window_min <= options.dt)
          return 1.0;
        else
          return 0.0;
      });

  double edge_count = static_cast<double>(network.edges_cause().size());
  std::vector<double> extenal_field_occupied_p;
  std::transform(
      in_comps.begin(), in_comps.end(),
      std::back_inserter(extenal_field_occupied_p),
      [&options, edge_count](const std::pair<EdgeType, CompType>& ic) {
        if (ic.second.cause_lifetime().first - options.window_min <= options.dt)
          return 1.0;
        else
          return ic.second.edge_set().estimate()/edge_count;
      });

  summary["dt-band-homogeneous-occupied"] = std::accumulate(
      in_comps.begin(), in_comps.end(), 0,
      [&options](size_t total, const std::pair<EdgeType, CompType>& ic) {
        if (ic.second.cause_lifetime().first - options.window_min <= options.dt &&
              options.window_max - ic.first.cause_time() <= options.dt)
          return total + 1;
        else
          return total;
      });

  summary["dt-band-external-field-occupied"] = std::accumulate(
      in_comps.begin(), in_comps.end(), 0.0,
      [&options, edge_count](double total, const std::pair<EdgeType, CompType>& ic) {
        if (options.window_max - ic.first.cause_time() > options.dt)
          return total;
        else if (ic.second.cause_lifetime().first - options.window_min <= options.dt)
          return total + 1.0;
        else
          return total + ic.second.edge_set().estimate()/edge_count;
      });

  summary["event-times-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
        event_times, event_times, time_bins,
        binned_math::count<typename EdgeType::TimeType>);
  summary["event-times-to-end-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, typename EdgeType::TimeType, size_t>(
        event_times_to_end, event_times_to_end, time_bins,
        binned_math::count<typename EdgeType::TimeType>);
  summary["homogeneous-occupied-count-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, double, size_t>(
        event_times, homogeneous_occupied_p, time_bins,
        binned_math::sum<double>);
  summary["homogeneous-occupied-count-to-end-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, double, size_t>(
        event_times_to_end, homogeneous_occupied_p, time_bins,
        binned_math::sum<double>);
  summary["external-field-occupied-count-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, double, size_t>(
        event_times, extenal_field_occupied_p, time_bins,
        binned_math::sum<double>);
  summary["external-field-occupied-count-to-end-dist"] =
    binned_math::binned_statistic<
      typename EdgeType::TimeType, double, size_t>(
        event_times_to_end, extenal_field_occupied_p, time_bins,
        binned_math::sum<double>);

  std::ofstream summary_file(options.summary_filename);
  summary_file << summary.dump() << std::endl;
}
