#include <string>
#include <fstream>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop
#include <nlohmann/json.hpp>

#include <dag/dag.hpp>
#include <hyperloglog.hpp>

#include "binned_math.tpp"

enum class prob_dist_types { deterministic, exponential };

template <class EdgeT>
struct opts {
  typename EdgeT::TimeType dt;

  bool provided_window_min;
  typename EdgeT::TimeType window_min;
  bool provided_window_max;
  typename EdgeT::TimeType window_max;

  size_t time_bins;

  size_t mass_bins;
  double mass_bins_min;

  size_t volume_bins;
  double volume_bins_min;

  size_t lifetime_bins;
  double lifetime_bins_min;

  bool logarithmic_bins;

  size_t seed;
  size_t temporal_reserve;

  std::string network_filename;
  std::string summary_filename;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("network_single",
        "Calculates binned out-component sizes on temporal network");
    option_def.add_options()
      ("seed", "seed for cardinality estimators",
      cxxopts::value<size_t>())
      ("dt", "delta-t value for this network",
      cxxopts::value<typename EdgeT::TimeType>())
      ("h,help", "Print help")
      ;

    option_def.add_options("Event List File")
      ("temporal-reserve", "estimated size of temporal network",
      cxxopts::value<size_t>()->default_value("0"))
      ("network", "network file",
      cxxopts::value<std::string>());

    option_def.add_options("Output")
      ("summary", "file to store binned out-component sizes of the network",
      cxxopts::value<std::string>())
      ("window-min", "start of measurement (default: minimum time of network)",
      cxxopts::value<typename EdgeT::TimeType>())
      ("window-max", "end of measurement (default: maximum time of network)",
      cxxopts::value<typename EdgeT::TimeType>())
      ("time-bins", "number of time bins (horizontal axis)",
      cxxopts::value<size_t>()->default_value("256"))
      ("logarithmic-bins", "use logarithmic bins for M, V and T "
      "quantities.")
      ("mass-bins", "number of mass bins",
      cxxopts::value<size_t>()->default_value("128"))
      ("mass-bins-min", "leftmost edge of mass bins",
      cxxopts::value<double>()->default_value("0"))
      ("volume-bins", "number of volume bins",
      cxxopts::value<size_t>()->default_value("128"))
      ("volume-bins-min", "leftmost edge of volume bins",
      cxxopts::value<double>()->default_value("0"))
      ("lifetime-bins", "number of lifetime bins",
      cxxopts::value<size_t>()->default_value("128"))
      ("lifetime-bins-min", "leftmost edge of lifetime bins",
      cxxopts::value<double>()->default_value("0"))
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("network") == 0) {
      std::cerr << "ERROR: needs a network argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("summary") == 0) {
      std::cerr << "ERROR: needs a summary file" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("dt") == 0) {
      std::cerr << "ERROR: needs a dt argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    }

    if (options.count("window-min") == 0) {
      provided_window_min = false;
    } else {
      provided_window_min = true;
      window_min = options["window-min"].as<typename EdgeT::TimeType>();
    }

    if (options.count("window-max") == 0) {
      provided_window_max = false;
    } else {
      provided_window_max = true;
      window_max = options["window-max"].as<typename EdgeT::TimeType>();
    }


    seed = options["seed"].as<size_t>();
    temporal_reserve = options["temporal-reserve"].as<size_t>();
    dt = options["dt"].as<typename EdgeT::TimeType>();

    time_bins = options["time-bins"].as<size_t>();

    mass_bins = options["mass-bins"].as<size_t>();
    mass_bins_min = options["mass-bins-min"].as<double>();

    volume_bins = options["volume-bins"].as<size_t>();
    volume_bins_min = options["volume-bins-min"].as<double>();

    lifetime_bins = options["lifetime-bins"].as<size_t>();
    lifetime_bins_min = options["lifetime-bins-min"].as<double>();

    network_filename = options["network"].as<std::string>();
    summary_filename = options["summary"].as<std::string>();

    logarithmic_bins = options["logarithmic-bins"].as<bool>();
  }

  void update_time_window(std::pair<typename EdgeT::TimeType, typename EdgeT::TimeType> time_window) {
    if (!provided_window_min)
      window_min = time_window.first;

    if (!provided_window_max)
      window_max = time_window.second;
  }
};

template <class EdgeT>
std::vector<EdgeT>
event_list(
    std::string net_filename,
    size_t temporal_reserve) {
  std::vector<EdgeT> topo;
  if (temporal_reserve > 0)
    topo.reserve(temporal_reserve);

  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}

#if defined(NETWORK_MODE_MOBILE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_TRANSPORT)
  using EdgeType = dag::directed_delayed_temporal_edge<uint32_t, uint32_t>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<12, 18>;
#elif defined(NETWORK_MODE_RANDOM_LARGE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED_LARGE)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
  using HllT = hll::HyperLogLog<10, 18>;
#else
  #error No network mode defined. should define one of the following:\
    `NETWORK_MODE_MOBILE`, `NETWORK_MODE_TRANSPORT`, `NETWORK_MODE_RANDOM`,\
    `NETWORK_MODE_RANDOM_LARGE`
#endif

using ProbType = dag::adjacency_prob::deterministic<EdgeType>;
template <class T>
using HLLEstimator = dag::hll_estimator<T, HllT>;

template <class T>
using HLLEstimatorRO = dag::hll_estimator_readonly<T, HllT>;
using CompType = dag::temporal_component<EdgeType, HLLEstimatorRO>;

int main(int argc, char* argv[]) {
  opts<EdgeType> options(argc, argv);

  // TODO: add exponential prob_dist_type

  auto events = event_list<EdgeType>(
      options.network_filename, options.temporal_reserve);

  dag::network<EdgeType> network(events);

  size_t eg_seed = 0ul;
  ProbType prob(options.dt);
  dag::implicit_event_graph<EdgeType, ProbType> eg(network, prob, eg_seed);

  options.update_time_window(eg.time_window());

  nlohmann::json summary = {
    {"seed", options.seed},
    {"dt", options.dt},
    {"window", {options.window_min, options.window_max}},
    {"prob_dist_type", "deterministic"},
  };

  summary["network"] = {
    {"filename", options.network_filename},
    {"temporal-edges", network.edges_cause().size()},
    {"temporal-nodes", network.vertices().size()},
  };

  std::vector<typename EdgeType::TimeType> time_bins =
    binned_math::linear_bins(
      options.window_min,
      options.window_max,
      options.time_bins);

  std::vector<double> mass_bins;
  std::vector<double> volume_bins;
  std::vector<double> lifetime_bins;

  if (options.logarithmic_bins) {
    mass_bins =
      binned_math::logarithmic_bins<double>(
        options.mass_bins_min, static_cast<double>(network.edges_cause().size()),
        options.mass_bins);
    volume_bins =
      binned_math::logarithmic_bins<double>(
        options.volume_bins_min, static_cast<double>(network.vertices().size()),
        options.volume_bins);
    lifetime_bins =
      binned_math::logarithmic_bins<double>(
        options.lifetime_bins_min, static_cast<double>(options.window_max-options.window_min),
        options.lifetime_bins);
  } else {
    mass_bins =
      binned_math::linear_bins<double>(
        options.mass_bins_min, static_cast<double>(network.edges_cause().size()),
        options.mass_bins);
    volume_bins =
      binned_math::linear_bins<double>(
        options.volume_bins_min, static_cast<double>(network.vertices().size()),
        options.volume_bins);
    lifetime_bins =
      binned_math::linear_bins<double>(
        options.lifetime_bins_min, static_cast<double>(options.window_max-options.window_min),
        options.lifetime_bins);
  }

  summary["time-bin-edges"] = time_bins;
  summary["mass-bin-edges"] = mass_bins;
  summary["volume-bin-edges"] = volume_bins;
  summary["lifetime-bin-edges"] = lifetime_bins;

  auto out_comps = dag::out_components<
    EdgeType, ProbType, HLLEstimator, HLLEstimatorRO>(
        eg, options.seed, false);

  std::sort(out_comps.begin(), out_comps.end(),
    [](auto const& a, auto const& b) { return a.first < b.first; });  // sort by cause-time


  std::vector<std::vector<size_t>> mass_bin_counts, volume_bin_counts, lifetime_bin_counts;

  mass_bin_counts.reserve(time_bins.size()-1ul);
  volume_bin_counts.reserve(time_bins.size()-1ul);
  lifetime_bin_counts.reserve(time_bins.size()-1ul);

  std::vector<double> current_bin_masses, current_bin_volumes, current_bin_lifetimes;

  auto current_comp = out_comps.begin();
  auto current_time_bin = time_bins.begin();
  while (current_comp != out_comps.end() && current_comp->first.cause_time() <= time_bins.back()) {
    if (current_comp->first.cause_time() >= *current_time_bin) {
      if (current_time_bin > time_bins.begin()) {
	mass_bin_counts.push_back(
          binned_math::binned_statistic<double, double, size_t>(
            current_bin_masses, current_bin_masses, mass_bins,
            binned_math::count<double>));
	volume_bin_counts.push_back( 
          binned_math::binned_statistic<double, double, size_t>(
            current_bin_volumes, current_bin_volumes, volume_bins,
            binned_math::count<double>));
	lifetime_bin_counts.push_back( 
          binned_math::binned_statistic<double, double, size_t>(
            current_bin_lifetimes, current_bin_lifetimes, lifetime_bins,
            binned_math::count<double>));
      }
      current_bin_masses.erase(current_bin_masses.begin(), current_bin_masses.end());
      current_bin_volumes.erase(current_bin_volumes.begin(), current_bin_volumes.end());
      current_bin_lifetimes.erase(current_bin_lifetimes.begin(), current_bin_lifetimes.end());
      current_time_bin++;
    }
    current_bin_masses.push_back(std::min(current_comp->second.edge_set().estimate(),
                                   static_cast<double>(mass_bins.back())));
    current_bin_volumes.push_back(std::min(current_comp->second.node_set().estimate(),
                                   static_cast<double>(volume_bins.back())));
    current_bin_lifetimes.push_back(static_cast<double>(
       current_comp->second.cause_lifetime().second - current_comp->second.cause_lifetime().first));
    current_comp++;
  }

  summary["mass-bin-counts"] = mass_bin_counts;
  summary["volume-bin-counts"] = volume_bin_counts;
  summary["lifetime-bin-counts"] = lifetime_bin_counts;

  std::ofstream summary_file(options.summary_filename);
  summary_file << summary.dump() << std::endl;
}
