#include <string>
#include <vector>
#include <fstream>
#include <algorithm>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop

#include <dag/dag.hpp>

enum class prob_dist_types { deterministic, exponential };

template <class EdgeT>
struct opts {
  typename EdgeT::TimeType dt;
  size_t seed;
  size_t sample_size;
  size_t temporal_reserve;
  typename EdgeT::TimeType time_res;

  std::string network_filename;
  std::string ocg_filename;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("network_spreading",
        "simulates spreading process on temporal networks");
    option_def.add_options()
      ("seed", "seed for random sampling",
      cxxopts::value<size_t>())
      ("dt", "delta-t value for this network",
      cxxopts::value<typename EdgeT::TimeType>())
      ("sample-size", "number of samples",
      cxxopts::value<size_t>()->default_value("1"))
      ("h,help", "Print help")
      ;

    option_def.add_options("Event List File")
      ("temporal-reserve", "estimated size of temporal network",
      cxxopts::value<size_t>()->default_value("0"))
      ("network", "network file",
      cxxopts::value<std::string>());

    option_def.add_options("Output")
      ("out-component-growth", "file to store out-component size vs. time",
      cxxopts::value<std::string>())
      ("time-res", "time resolution for out-component size file",
      cxxopts::value<typename EdgeT::TimeType>()->default_value("1"))
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("dt") == 0) {
      std::cerr << "ERROR: needs a dt argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("network") == 0) {
      std::cerr << "ERROR: needs a network argument" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    } else if (options.count("out-component-growth") == 0) {
      std::cerr << "ERROR: needs a out-component-growth file" << "\n";
      std::cerr << option_def.help({"", "Event List File", "Output"})
        << "\n";
      std::exit(1);
    }


    seed = options["seed"].as<size_t>();
    network_filename = options["network"].as<std::string>();
    temporal_reserve = options["temporal-reserve"].as<size_t>();
    sample_size = options["sample-size"].as<size_t>();
    dt = options["dt"].as<typename EdgeT::TimeType>();
    time_res = options["time-res"].as<typename EdgeT::TimeType>();
    ocg_filename = options["out-component-growth"].as<std::string>();
  }
};


template <class EdgeT>
std::vector<EdgeT>
event_list(
    std::string net_filename,
    size_t temporal_reserve) {

  std::vector<EdgeT> topo;
  if (temporal_reserve > 0)
    topo.reserve(temporal_reserve);

  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}

#if defined(NETWORK_MODE_MOBILE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, uint32_t>;
#elif defined(NETWORK_MODE_TRANSPORT)
  using EdgeType = dag::directed_delayed_temporal_edge<uint32_t, uint32_t>;
#elif defined(NETWORK_MODE_RANDOM)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
#elif defined(NETWORK_MODE_RANDOM_LARGE)
  using EdgeType = dag::undirected_temporal_edge<uint32_t, double>;
#elif defined(NETWORK_MODE_RANDOM_DIRECTED_LARGE)
  using EdgeType = dag::directed_temporal_edge<uint32_t, double>;
#else
  #error No network mode defined. should define one of the following: \
    `NETWORK_MODE_MOBILE`, `NETWORK_MODE_TRANSPORT`, `NETWORK_MODE_RANDOM`, \
    `NETWORK_MODE_RANDOM_LARGE`
#endif

using ProbType = dag::adjacency_prob::deterministic<EdgeType>;

int main(int argc, char* argv[]) {
  opts<EdgeType> options(argc, argv);

  std::ofstream ocg_file(options.ocg_filename);

  ProbType prob(options.dt);
  size_t eg_seed = 0ul;

  dag::implicit_event_graph<EdgeType, ProbType> eg(
      dag::network<EdgeType>(
        event_list<EdgeType>(
          options.network_filename, options.temporal_reserve)),
      prob, eg_seed);

  size_t estimator_seed = 0ul;

  std::vector<EdgeType> event_samples;
  std::mt19937 gen(options.seed);
  std::sample(
      eg.events_cause().begin(),
      eg.events_cause().end(),
      std::back_inserter(event_samples),
      options.sample_size,
      gen
      );

  std::vector<std::pair<size_t, size_t>> counts;
  counts.emplace_back(0ul, 0ul);

  for (auto&& root: event_samples) {
    auto comp_exact = dag::out_component<
      EdgeType, ProbType, dag::exact_estimator>(
          eg, root,
          estimator_seed,
          0ul, 0ul);

    std::vector<EdgeType> events(
        comp_exact.edge_set().set().begin(),
        comp_exact.edge_set().set().end());

    std::sort(events.begin(), events.end()); // sort by cause time

    comp_exact = dag::temporal_component<EdgeType, dag::exact_estimator>(
        estimator_seed,
        comp_exact.node_set().size(),
        comp_exact.edge_set().size());

    std::vector<std::pair<size_t, size_t>> root_counts;
    root_counts.emplace_back(0ul, 0ul);

    for (auto&& e: events) {
      comp_exact.insert(e, e.mutated_verts());
      size_t bin = static_cast<size_t>(
          static_cast<double>(
            e.cause_time()-root.cause_time())/options.time_res);

      while (root_counts.size() <= bin)
        root_counts.push_back(root_counts.back());

      root_counts[bin] = std::make_pair(
            comp_exact.node_set().size(),
            comp_exact.edge_set().size());
    }

    size_t max_size = std::max(root_counts.size(), counts.size());
    while (root_counts.size() < max_size)
      root_counts.push_back(root_counts.back());
    while (counts.size() < max_size)
      counts.push_back(counts.back());

    std::transform(
        root_counts.begin(), root_counts.end(),
        counts.begin(),
        counts.begin(),
        [](std::pair<size_t, size_t> a, std::pair<size_t, size_t> b) {
          return std::make_pair(a.first + b.first, a.second + b.second);
        });
  }

  EdgeType::TimeType t = 0;
  double sample_size = static_cast<double>(event_samples.size());
  for (const auto& [node_count, event_count]: counts) {
    ocg_file
      << t << " "
      << static_cast<double>(node_count)/sample_size << " "
      << static_cast<double>(event_count)/sample_size << "\n";
    t += options.time_res;
  }
}
