#include <iomanip>
#include <string>
#include <fstream>
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#include <cxxopts.hpp>
#pragma GCC diagnostic pop

#include <dag/dag.hpp>
#include <hyperloglog.hpp>

enum class static_models{ erdos, barabasi, grid, regular };
enum class temporal_models { poisson, bursty, self_exciting };

template <class RealType = double>
class power_law_with_specified_mean {
private:
  RealType x_min;
public:
  RealType exponent, mean;

  power_law_with_specified_mean(RealType exponent, RealType mean) :
  exponent(exponent), mean(mean) {
    if (exponent <= 2)
      throw std::domain_error("exponent should be larger than 2");
    if (mean <= 0)
      throw std::domain_error("mean should be positive");

   x_min = mean*(exponent - 2.0)/(exponent - 1.0);
  }

  template <class Generator>
  RealType operator()(Generator& g) {
    std::uniform_real_distribution<RealType> dist;
    return x_min*std::pow((1.0 - dist(g)), 1.0/(1.0 - exponent));
  }
};

template <class RealType = double>
class residual_power_law_with_specified_mean {
private:
  RealType x_min;
public:
  RealType exponent, mean;

  residual_power_law_with_specified_mean(RealType exponent, RealType mean) :
  exponent(exponent), mean(mean) {
    if (exponent <= 2)
      throw std::domain_error("exponent should be larger than 2");
    if (mean <= 0)
      throw std::domain_error("mean should be positive");

    x_min = mean*(exponent - 2.0)/(exponent - 1.0);
  }

  template <class Generator>
  RealType operator()(Generator& g) {
    RealType p = std::uniform_real_distribution<RealType>()(g);
    if (p < (exponent - 2.0)/(exponent - 1.0))
      return (exponent - 1.0)/(exponent - 2.0)*x_min*p;
    else
      return x_min*std::pow((exponent - 1.0)*(1.0 - p), 1.0/(2.0 - exponent));
  }
};

class hawkes_process {
public:
  double mu;
  double alpha;
  double theta;
  double phi;

  hawkes_process(double mu, double alpha, double theta, double phi=0.0):
    mu(mu), alpha(alpha), theta(theta), phi(phi) {}

  template <class Generator>
  double operator()(Generator& g) {
    // Ogata’s modified thinning algorithm.
    // inspired by https://github.com/canerturkmen/hawkeslib/blob/37f7b0c247568bb5ae96eb83d89a2d9bbbc4d43c/hawkeslib/model/c/c_uv_exp.pyx#L167
    double iet = 0.0, d = 0.0, ed, m, lda, r2;

    do {
      m = mu + alpha * theta * (1.0 + phi);

      double e = -std::log(std::uniform_real_distribution()(g))/m;
      iet += e;

      ed = std::exp(-theta * (e + d));
      lda = mu + alpha * theta * ed * (1.0 + phi);

      r2 = std::uniform_real_distribution()(g);
      d += e;
    } while (r2 * m > lda);

    phi = ed * (1.0 + phi);

    return iet;
  }
};

template <class EdgeT>
struct opts {
  size_t seed;

  // temporal_model
  // poisson + bursty
  typename EdgeT::TimeType mean_dt;
  typename EdgeT::TimeType max_time;
  typename EdgeT::TimeType exponent;

  // self_exciting
  double background_intensity;
  double infectivity;
  double exponential_delay_density_rate;

  // static_model
  typename EdgeT::VertexType nodes;
  size_t dims;
  double average_degree;
  bool directed;

  static_models static_model;
  std::string static_path;
  temporal_models temporal_model;

  opts(int argc, char* argv[]) {
    cxxopts::Options option_def("random_network",
        "Generates random temporal networks");
    option_def.add_options()
      ("seed", "seed for pseudo-random number generator",
      cxxopts::value<size_t>())
      ("nodes", "number of nodes in the base network",
      cxxopts::value<typename EdgeT::VertexType>())
      ("average-degree", "average degree in the base network",
      cxxopts::value<double>())
      ("dims", "number of dimensions of the network (square grid)",
      cxxopts::value<size_t>())
      ("static-model", "Base static network model used. acceptable inputs:"
       "`erdos` for Erdős–Rényi (default), `barabasi` for Barabási–Albert, "
       "`grid` for square grids with periodic boundary condition, `regular`"
       " for random k-regular network or "
       "path to an edgelist file",
      cxxopts::value<std::string>()->default_value("erdos"))
      ("directed", "create two directed edges for each edge to avoid re-infection",
      cxxopts::value<bool>()->default_value("false"))
      ("temporal-model", "Inter-event time distributions. acceptable inputs:"
       "`poisson` for poisson process (exponential inter-event times, default) "
       "`self-exciting` for Hawkes process and `bursty` for power law with cutoff.",
      cxxopts::value<std::string>()->default_value("poisson"))
      ("mean-dt", "expected value of inter-event time distribution",
      cxxopts::value<typename EdgeT::TimeType>())
      ("exponent", "exponent for bursty inter-event time. should be larger than 2.",
      cxxopts::value<typename EdgeT::TimeType>())
      ("max-time", "maximum time for observation window",
      cxxopts::value<typename EdgeT::TimeType>())
      ("background-intensity", "background rate of exogenous events `mu`",
      cxxopts::value<double>())
      ("infectivity", "expected number of induced events or \"aftershocks\" `alpha`",
      cxxopts::value<double>())
      ("exponential-delay-density-rate", "rate of delay between exogenous events and their"
       " induced events `theta`",
      cxxopts::value<double>())
      ("h,help", "Print help")
      ;

    auto options = option_def.parse(argc, argv);

    if (options.count("help") != 0) {
      std::cerr << option_def.help({""})
        << std::endl;
      std::exit(0);
    } else if (options.count("seed") == 0) {
      std::cerr << "ERROR: needs a seed argument" << std::endl;
      std::cerr << option_def.help({""})
        << std::endl;
      std::exit(1);
    } else if (options.count("nodes") == 0) {
      std::cerr << "ERROR: needs a node argument" << std::endl;
      std::cerr << option_def.help({""})
        << std::endl;
      std::exit(1);
    } else if (options.count("max-time") == 0) {
      std::cerr << "ERROR: needs a max-time argument" << std::endl;
      std::cerr << option_def.help({""})
        << std::endl;
      std::exit(1);
    } 

    seed = options["seed"].as<size_t>();
    nodes = options["nodes"].as<typename EdgeT::VertexType>();
    max_time = options["max-time"].as<typename EdgeT::TimeType>();
    directed = options["directed"].as<bool>();

    if (options["static-model"].as<std::string>() == "erdos") {
      if (options.count("average-degree") == 0) {
        std::cerr << "ERROR: `erdos` network needs an average-degree argument" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      }
      average_degree = options["average-degree"].as<double>();
      static_model = static_models::erdos;
    } else if (options["static-model"].as<std::string>() == "barabasi") {
      if (options.count("average-degree") == 0) {
        std::cerr << "ERROR: `barabasi` network needs an average-degree argument" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      }
      average_degree = options["average-degree"].as<double>();
      static_model = static_models::barabasi;
    } else if (options["static-model"].as<std::string>() == "regular") {
      if (options.count("average-degree") == 0) {
        std::cerr << "ERROR: `regular` network needs an average-degree argument" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      }
      average_degree = options["average-degree"].as<double>();
      static_model = static_models::regular;
    } else if (options["static-model"].as<std::string>() == "grid") {
      if (options.count("dims") == 0) {
        std::cerr << "ERROR: `grid` network needs an dimensions argument" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      }
      dims = options["dims"].as<size_t>();
      static_model = static_models::grid;
    } else {
      static_path = options["static-model"].as<std::string>();
    }

    if (options["temporal-model"].as<std::string>() == "poisson") {
      temporal_model = temporal_models::poisson;
      if (options.count("mean-dt") == 0) {
        std::cerr << "ERROR: poisson temporal-model needs a mean-dt argument" << std::endl;
        std::cerr << option_def.help({""})
          << std::endl;
        std::exit(1);
      } else {
        mean_dt = options["mean-dt"].as<typename EdgeT::TimeType>();
      }
    } else if (options["temporal-model"].as<std::string>() == "bursty") {
      temporal_model = temporal_models::bursty;

      if (options.count("exponent") == 0) {
        std::cerr << "ERROR: bursty temporal-model needs an exponent" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      } else if (options.count("mean-dt") == 0) {
        std::cerr << "ERROR: bursty temporal-model needs a mean-dt argument" << std::endl;
        std::cerr << option_def.help({""})
          << std::endl;
        std::exit(1);
      } else {
        exponent = options["exponent"].as<typename EdgeT::TimeType>();
        mean_dt = options["mean-dt"].as<typename EdgeT::TimeType>();
      }
    } else if (options["temporal-model"].as<std::string>() == "self-exciting") {
      temporal_model = temporal_models::self_exciting;
      if (options.count("background-intensity") == 0) {
        std::cerr << "ERROR: self_exciting temporal-model needs a background-intensity" << std::endl;
        std::cerr << option_def.help({""}) << std::endl;
        std::exit(1);
      } else if (options.count("infectivity") == 0) {
        std::cerr << "ERROR: self_exciting temporal-model needs an infectivity" << std::endl;
        std::cerr << option_def.help({""})
          << std::endl;
        std::exit(1);
      } else if (options.count("exponential-delay-density-rate") == 0) {
        std::cerr << "ERROR: self_exciting temporal-model needs an exponential-delay-density-rate" << std::endl;
        std::cerr << option_def.help({""})
          << std::endl;
        std::exit(1);
      } else  {
        background_intensity = options["background-intensity"].as<double>();
        infectivity = options["infectivity"].as<double>();
        exponential_delay_density_rate = options["exponential-delay-density-rate"].as<double>();
      }
    } else {
      throw std::invalid_argument("bad value for argument `temporal-model`");
    }
  }
};

template <class EdgeT>
std::vector<EdgeT>
edge_list(std::string net_filename) {
  std::vector<EdgeT> topo;
  std::ifstream net(net_filename);

  // self links should be removed
  EdgeT e;
  while (net >> e)
    if (e.mutator_verts() != e.mutated_verts() ||
        e.mutated_verts().size() > 1)
      topo.push_back(e);

  std::sort(topo.begin(), topo.end());
  topo.erase(std::unique(topo.begin(), topo.end()) , topo.end());

  return topo;
}


template <class VertT>
dag::undirected_network<VertT> path_graph(VertT size, bool periodic=false) {
  std::vector<dag::undirected_edge<VertT>> edges;
  edges.reserve(size);

  for (VertT i = {}; i < size-1; i++)
    edges.emplace_back(i, i+1);

  if (periodic)
    edges.emplace_back(size - 1, 0);

  return dag::undirected_network<VertT>(edges);
}

template <class VertT1, class VertT2>
dag::undirected_network<std::pair<VertT1, VertT2>>
cartesian_product(dag::undirected_network<VertT1> g1, dag::undirected_network<VertT2> g2) {
  std::vector<dag::undirected_edge<std::pair<VertT1, VertT2>>> edges;
  edges.reserve(g1.vertices().size()*g2.edges().size() + g1.edges().size()*g2.vertices().size());

  for (auto&& e: g1.edges()) {
     auto verts = e.mutator_verts();
     if (verts.size() > 1)
       for (auto&& n: g2.vertices())
         edges.emplace_back(std::make_pair(verts[0], n), std::make_pair(verts[1], n));
  }

  for (auto&& n: g1.vertices()) {
    for (auto&& e: g2.edges()) {
     auto verts = e.mutator_verts();
     if (verts.size() > 1)
       edges.emplace_back(std::make_pair(n, verts[0]), std::make_pair(n, verts[1]));
    }
  }

  return dag::undirected_network<std::pair<VertT1, VertT2>>(edges);
}

namespace std {
  template<typename T1, typename T2>
  struct hash<std::pair<T1, T2>> {
    size_t
    operator()(const std::pair<T1, T2>& p) const {
      return dag::utils::combine_hash(std::hash<T1>{}(p.first), p.second);
    }
  };
}

template <class OutVertT, class InVertT>
dag::undirected_network<OutVertT>
relabel_nodes(dag::undirected_network<InVertT> g) {
  std::unordered_map<InVertT, OutVertT> new_labels;
  new_labels.reserve(g.vertices().size());

  OutVertT i = {}; 
  for (auto&& n: g.vertices())
    new_labels.emplace(n, i++);

  std::vector<dag::undirected_edge<OutVertT>> edges;
  for (auto&& e: g.edges()) {
    auto verts = e.mutator_verts();
    if (verts.size() > 1)
      edges.emplace_back(new_labels.at(verts[0]), new_labels.at(verts[1]));
  }

  return dag::undirected_network<OutVertT>(edges);
}

template <class VertT>
dag::undirected_network<VertT> square_grid_graph(VertT side, size_t dims, bool periodic=false) {
  auto g = path_graph<VertT>(side, periodic);
  for (size_t i = 1; i < dims; i++)
    g = relabel_nodes<VertT>(cartesian_product(g, path_graph<VertT>(side, periodic)));

  return g;
}



template <class VertT>
bool _random_regular_suitable_hanging_edge(
  const std::unordered_set<dag::undirected_edge<VertT>>& edges,
  const std::unordered_map<VertT, size_t>& hanging_stubs) {
  if (hanging_stubs.empty())
    return true;
  else
    for (auto& [node1, count1]: hanging_stubs)
      for (auto& [node2, count1]: hanging_stubs)
        if (node1 == node2)
          break;
        else if (edges.find({node1, node2}) == edges.end())
          return true;

  return false;
}

template <class VertT>
dag::undirected_network<VertT> random_regular_graph(size_t size, size_t degree, std::mt19937_64& gen) {
  if (size*degree % 2 != 0)
    throw std::domain_error("size * degree must be even");

  if (degree >= size)
    throw std::domain_error("degree must be less than size");

  // There is always an answer. We Just have to look hard enough. 
  while (true) {
    std::unordered_set<dag::undirected_edge<VertT>> edges;
    std::vector<VertT> stubs;
    stubs.reserve(size*degree);
    for (size_t i = 0; i < size; i++)
      for (size_t d = 0; d < degree; d++)
        stubs.emplace_back(i);

    bool suitable = true;
    while (!stubs.empty() && suitable) {
      std::shuffle(stubs.begin(), stubs.end(), gen);
      std::unordered_map<VertT, size_t> hanging_stubs;
      for (auto iter = stubs.begin(); iter < stubs.end(); std::advance(iter, 2)) {
        if (*iter != *(iter+1) && edges.find({*iter, *(iter+1)}) == edges.end()) {
          edges.emplace(*iter, *(iter+1));
        } else {
          hanging_stubs[*iter]++;
          hanging_stubs[*(iter+1)]++;
	}
      }

      suitable = _random_regular_suitable_hanging_edge(edges, hanging_stubs);

      stubs.clear();
      for (auto& [node, count]: hanging_stubs)
        for (size_t i = 0; i < count; i++)
           stubs.emplace_back(node);
    }

    if (suitable)
      return dag::undirected_network<VertT>(std::vector<dag::undirected_edge<VertT>>(edges.begin(), edges.end()));
  }

  // Unreachable.
  return dag::undirected_network<VertT>();
}

int main(int argc, char* argv[]) {
  using VertexType = uint32_t;
  using TimeType = double;
  using StaticEdgeType = dag::undirected_edge<VertexType>;
  using EdgeType = dag::undirected_temporal_edge<VertexType, TimeType>;

  opts<EdgeType> options(argc, argv);

  dag::network<StaticEdgeType> base;
  std::mt19937_64 gen(options.seed);

  if (!options.static_path.empty()) {
    base = dag::undirected_network<VertexType>(
        edge_list<dag::undirected_edge<VertexType>>(options.static_path));
  } else if (options.static_model == static_models::erdos) {
    base = dag::gnp_random_graph<VertexType>(
        options.nodes,
        options.average_degree/static_cast<double>(options.nodes-1),
        gen);
  } else if (options.static_model == static_models::barabasi) {
    base = dag::ba_random_graph<VertexType>(
        options.nodes,
        std::lround(options.average_degree/2.0),
        gen);
  } else if (options.static_model == static_models::grid) {
    double side = std::pow(static_cast<double>(options.nodes), 1.0/static_cast<double>(options.dims));
    if (std::abs(side - std::round(side)) > 1e-3/side)
      throw std::domain_error("square grid nodes should be a power of dims");
    base = square_grid_graph<VertexType>(
        static_cast<VertexType>(std::llround(side)),
        options.dims,
        true);
  } else if (options.static_model == static_models::regular) {
    base = random_regular_graph<VertexType>(
        options.nodes,
        static_cast<size_t>(options.average_degree),
        gen);
  }

  std::vector<EdgeType> edges;
  if (options.temporal_model == temporal_models::poisson) {
    std::exponential_distribution<TimeType> dist(options.mean_dt);
    edges = dag::random_events<EdgeType,
          std::exponential_distribution<TimeType>>(
        base,
        options.max_time,
        dist, dist,
        options.seed);
  } else if (options.temporal_model == temporal_models::bursty) {
    power_law_with_specified_mean<TimeType> dist(options.exponent, options.mean_dt);
    residual_power_law_with_specified_mean<TimeType> res_dist(options.exponent, options.mean_dt);

    edges = dag::random_events<EdgeType,
          power_law_with_specified_mean<TimeType>,
          residual_power_law_with_specified_mean<TimeType>>(
        base,
        options.max_time,
        dist, res_dist,
        options.seed);
  } else if (options.temporal_model == temporal_models::self_exciting) {
    hawkes_process dist(options.background_intensity, options.infectivity, options.exponential_delay_density_rate);

    edges = dag::random_events<EdgeType, hawkes_process>(
        base,
        options.max_time,
        dist,
        options.seed);
  }

  std::sort(edges.begin(), edges.end());
  std::cout << std::setprecision(std::numeric_limits<TimeType>::digits10 + 1);
  for (auto& e: edges) {
    auto verts = e.mutator_verts();
    for (auto& v: verts)
      std::cout << v << " ";
    std::cout << e.effect_time() << "\n";

    if (options.directed) {
      std::reverse(verts.begin(), verts.end());
      for (auto& v: verts)
        std::cout << v << " ";
      std::cout << e.effect_time() << "\n";
    }
  }
}

